/*
** Copyright (C) 2000 George Tzanetakis <gtzan@cs.princeton.edu>
**  
** This program is free software; you can redistribute it and/or modify
** it under the terms of the GNU General Public License as published by
** the Free Software Foundation; either version 2 of the License, or
** (at your option) any later version.
** 
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU General Public License for more details.
** 
** You should have received a copy of the GNU General Public License
** along with this program; if not, write to the Free Software 
** Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
*/


/**
   \class  BeatExtractor
   \brief  Single Vector feature extractor for beat information

   Feature vector of beat related information calculated using the DWT. 
*/



#include "BeatExtractor.h"

BeatExtractor::BeatExtractor(MSignal* in):Extractor(in)
{
  unsigned int i;
  unsigned int inSize = in->winSize();
  
  in_.create(inSize);
  dwtres_.create(inSize);
  sum_.create(inSize);
  tmp_.create(inSize);
  band_.create(inSize);
  lpfband_.create(inSize);
  rectband_.create(inSize);
  normband_.create(inSize);
  
  normRMS_ = new NormRMS(inSize);

  daub4_ = new Daub4(inSize, WVL_FORWARD);
  idaub4_ = new Daub4(inSize, WVL_INVERSE);
  pyramid_ = new WaveletPyramid(inSize, daub4_, WVL_FORWARD);
  ipyramid_ = new WaveletPyramid(inSize, idaub4_, WVL_INVERSE);

  rect_ = new FullWaveRectifier(inSize);
  lpf_ = new OnePole(inSize, 0.99);
  norm_ = new Norm(inSize);
  
  dsfactor_ = 16;
  ds_ = new DownSampler(inSize, dsfactor_);
  dsampled_.create(inSize / dsfactor_);  
  
  autocor2_ = new AutoCorrelationFFT(inSize / dsfactor_);
  autoc_.create(inSize / dsfactor_);
  
  /* AutoCorrelation autocor1(src->winSize()/factor); */
  pe2_ = new PeakEnhancer(inSize/dsfactor_, 2);
  pe3_ = new PeakEnhancer(inSize/dsfactor_, 3);
  pe4_ = new PeakEnhancer(inSize/dsfactor_, 4);
  picker_ = new PeakPicker(inSize, dsfactor_, 3, 300, 3200, 100);
  peaks_.create(picker_->outSize());
  
  featSize_ = 6;
  out_.create(1);
  res_.create(featSize_);

  histo_.create(300);
  int k;
  
  for (k=0; k < featSize_; k++)
    featNames_.push_back("bf");
}

BeatExtractor::~BeatExtractor()
{
  delete daub4_;
  delete idaub4_;
  delete pyramid_;
  delete ipyramid_;
  delete normRMS_;
  delete rect_;
  delete lpf_;
  delete norm_;
  delete ds_;
  delete autocor2_;
  delete picker_;
  delete pe2_;
  delete pe3_;
  delete pe4_;  
}

unsigned int 
BeatExtractor::outSize() 
{
  return featSize_;
}


void
BeatExtractor::extract()
{
  extract(0, src_->iterations());
}

void 
BeatExtractor::extract(unsigned int start, unsigned int end, 
		      FeatMatrix& fmatrix) 
{
  int load_matrix = 1;
  unsigned int t;
  unsigned int i;

  unsigned int base = 4;
  unsigned int hlevel;
  unsigned int llevel;
  unsigned int level;
  
  if ((fmatrix.rows() != 1) || (fmatrix.cols() != res_.size()))
    {
      cerr << "BeatExtractor::extract incorrect matrix dimensions " << endl;
      cerr << "Rows = " << fmatrix.rows() << endl;
      cerr << "Cols = " << fmatrix.cols() << endl;
      cerr << "out_.size = " << out_.size() << endl;
      load_matrix = 0;
    }
  fmatrix.setFeatNames(featNames_);
  unsigned int p=0;
  histo_.setval(0.0);
  
  for (t=start; t< end; t++)
    {
      sum_.setval(0.0);
      src_->get(in_, t);  
      //normRMS_->process(in_, in_);
      pyramid_->process(in_, dwtres_);
      
      for (i=0; i < 5; i++)
	{
	  level = 7 + i;
	  hlevel = base << level;
	  llevel = base << (level-1);
	  tmp_ = dwtres_;
	  tmp_.setval(hlevel, dwtres_.size(), 0.0);
	  tmp_.setval(0, llevel, 0.0);
	  ipyramid_->process(tmp_, band_);
	  rect_->process(band_, rectband_);
	  lpf_->process(rectband_, lpfband_);
	  norm_->process(rectband_, normband_);
	  sum_ += normband_;
	}
      ds_->process(sum_, dsampled_);
      autocor2_->process(dsampled_, autoc_);
      picker_->process(autoc_, peaks_);

      // convert peaks to histogram bins
      float pitch;
      unsigned int bin;
      float amp;
      float mpos;
      for (i=0; i < peaks_.size()/2; i++)
	{
	  mpos = peaks_(i + peaks_.size()/2);
	  if (mpos > 0.0)	  
	    {
	      pitch=22050.0/ (mpos  * dsfactor_);	     
	      bin = ((unsigned int)(pitch * 60.0));
	      // amp = 1.0;
	      amp = peaks_(i);
	      assert(bin < histo_.size());
	      histo_(bin) += amp;
	    }
	}
      if (t % ((end-start)/5) == 0)
	{
	  cerr << p * 20  << "% ";
	  p++;
	}      
    }
  
  cerr << endl;
  cerr << "BeatExtractor::extract completed" << endl;
  // Compute features from histogram 
  BeatHistoFeatures bhf(histo_.size());
  bhf.process(histo_, res_);
  
  fmatrix.row(0, res_);
}


void
BeatExtractor::extract(unsigned int start, unsigned int end)
{
  unsigned int t;
  for (t=start; t< end; t++)
    {
      src_->get(in_, t);
      if (t % ((end-start)/10))
	{
	  cerr << (end-start)/t * 100.0 << "% processed" << endl;
	}      
    }
}

