/*
** Copyright (C) 2000 George Tzanetakis <gtzan@cs.princeton.edu>
**  
** This program is free software; you can redistribute it and/or modify
** it under the terms of the GNU General Public License as published by
** the Free Software Foundation; either version 2 of the License, or
** (at your option) any later version.
** 
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU General Public License for more details.
** 
** You should have received a copy of the GNU General Public License
** along with this program; if not, write to the Free Software 
** Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
*/

/** 
    \class Classifier.h
    \brief Base class for classifiers

    Base class that all classifiers inherit. Specifies the generic
    interface any classifier will have and implements the methods that
    are independent of the classification method.  
*/



#if !defined(__Classifier_h)
#define __Classifier_h

#include <stdio.h> 
#include <string.h>
using namespace std;
#include "fvec.h"
#include "ivec.h"
#include "FeatMatrix.h"
#include "defs.h" 

class Classifier
{
protected:
  unsigned int classSize_;
  vector<string> classNames_;
  unsigned int featSize_;
  vector<string> featNames_;
  string type_;
  string name_;
  FeatMatrix* trainMatrix_;
  ivec    mask_;
  unsigned int lcol_;
  
  
public:
  Classifier();
  virtual ~Classifier();
  inline float mahalanobis(fvec& vec1, const fvec& vec2, const fvec& covars);
  virtual void train(FeatMatrix* fmatrix, unsigned int lcol) = 0;
  virtual unsigned int classify(fvec& vec) = 0;
  string estimate(FeatMatrix& matrix, unsigned int start, 
			unsigned int end);
  string estimate(FeatMatrix& matrix);
  unsigned int estimateClass(FeatMatrix& matrix, unsigned int start, 
			     unsigned int end);
  unsigned int estimateClass(FeatMatrix& matrix);
  virtual float evaluate(FeatMatrix* matrix);
  void classifyFeatMatrix(FeatMatrix& matrix);
  string className(unsigned int i);
  unsigned int classSize();
  void confusion(FeatMatrix* matrix, fmatrix& confusionMatrix);
  vector<string> rankList(FeatMatrix& qmap, unsigned int listSize);  
  void mask(unsigned int classNum);

};


inline float 
Classifier::mahalanobis(fvec& vec1, const fvec& vec2, const fvec& covars)
{
  unsigned int i;
  float sq_sum = 0.0;
  unsigned int size = vec1.size();
  float diff;
  for (i=0; i<size; i++)
    {
      diff = vec1(i) - vec2(i);
      sq_sum += (diff * covars(i) * diff);
    }

  return sq_sum;
}


  
#endif
