 
/*
** Copyright (C) 2000 George Tzanetakis <gtzan@cs.princeton.edu>
**  
** This program is free software; you can redistribute it and/or modify
** it under the terms of the GNU General Public License as published by
** the Free Software Foundation; either version 2 of the License, or
** (at your option) any later version.
** 
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU General Public License for more details.
** 
** You should have received a copy of the GNU General Public License
** along with this program; if not, write to the Free Software 
** Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
*/




/**  
     \class Filter
     \brief Filter a MSignal 

     Filters a MSignal using the coefficients stored in a
FilterCoeffs object. The filter is implemented as a direct form II
structure. This is a canonical form which has the minimum number of
delay elements.  
*/



#include "Filter.h"



Filter::Filter(fvec ncoeffs, fvec dcoeffs, string type):System()
{
  int i;
  norder_ = ncoeffs.size();
  dorder_ = dcoeffs.size();
  ncoeffs_.allocate(norder_);
  dcoeffs_.allocate(dorder_);
  ncoeffs_ = ncoeffs;
  dcoeffs_ = dcoeffs;
  type_ = type;
  if (norder_ > dorder_)
    order_ = norder_;
  else 
    order_ = dorder_;


  unsigned int stateSize = order_-1;
  state_.create(stateSize);
  stateSize_ = stateSize;


  
  if (dcoeffs_(0) != 1.0) 
    {
      for (i=0; i< (int)dorder_; i++)
	{
	  dcoeffs_(i) /= dcoeffs_(0);
	}
      
      for (i=0; i< (int)norder_; i++)
	{
	  ncoeffs_(i) /= dcoeffs_(0);
	}
    }
  fgain_ = 1.0;
  
}



Filter::~Filter()
{
}

unsigned int 
Filter::norder()
{
  return norder_;
}

unsigned int 
Filter::dorder()
{
  return dorder_;
}


unsigned int 
Filter::order()
{
  return order_;
}


void 
Filter::write(string filename)
{
  ofstream os(filename.c_str());
  os << (*this) << endl;
}


void 
Filter::read(string filename)
{
  ifstream is(filename.c_str());
  is >> (*this);
}


ostream& operator << (ostream& o, const Filter& filter)
{
  o << "# MARSYAS Filter \n";
  o << filter.ncoeffs_;
  o << filter.dcoeffs_;
  o << "# MARSYAS Filter \n";
  return o;
}


istream& operator >> (istream& i, Filter& filter)
{
  string str0, str1, str2;
  i >> str0 >> str1 >> str2;
  i >> filter.ncoeffs_;
  i >> filter.dcoeffs_;
  i >> str0 >> str1 >> str2;
  return i;
}


 
void
Filter::process(fvec& in, fvec& out)
{
  int i,j;
  int size = in.size();
  
  if (out.size() != in.size())
    {
      cerr << "Filter::process in and out must have the same size\n";
      return;
    }
  
  // State array holds the various delays for the difference equation 
  // of the filter. Similar implementation as described in the manual
  // for MATLAB MSignal Processing Toolbox. State corresponds to 
  // the z, num_coefs to the b and denom_coefs to the a vector respectively
  // in_window is the input x(n) and out_window is the output y(n)
  
  for (i=0; i<size; i++)
    {
      out(i) = (ncoeffs_(0) * in(i) + state_(0));
      
      for (j=0; j< stateSize_-1; j++)
	{
	  state_(j) = ncoeffs_(j+1) * in(i) + state_(j+1) 
	    - dcoeffs_(j+1) * out(i);
	  if (j==0)
	    {
	      ;
	      //cout << "state 0 = " << state_(j) << endl;
	    }
	}
      state_(stateSize_-1) = ncoeffs_(order_-1) * in(i) -
	dcoeffs_(order_-1) * out(i);
    }
  out *= fgain_;
}




