/*
** Copyright (C) 2000 George Tzanetakis <gtzan@cs.princeton.edu>
**  
** This program is free software; you can redistribute it and/or modify
** it under the terms of the GNU General Public License as published by
** the Free Software Foundation; either version 2 of the License, or
** (at your option) any later version.
** 
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU General Public License for more details.
** 
** You should have received a copy of the GNU General Public License
** along with this program; if not, write to the Free Software 
** Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
*/


/** 
    \class GMM
    \brief GMM classifier

    GMM (Gaussian Mixture Model) classifier is a classifier where each
class is modelled as a mixture of multidimensional Gaussian
distribution. The features are assumed to be uncorrelated and a
diagonal covariance matrix is used.  */





#include "GMM.h"

GMM::GMM()
{
  type_ = "GMM";
  name_ = "DefaultName";
  k_ = 0;
  lcol_ = 0;
}


GMM::GMM(unsigned int k)
{
  type_ = "GMM";
  name_ = "DefaultName";
  k_ = k;
  lcol_ = 0;
  iterations_ = 200;			// GMM iterations
  kiterations_ = 100;			// K-MEANS initialization iterations
  eiterations_ = 20;			// evaluation iterations
}

GMM::~GMM()
{
}

void 
GMM::mask(ivec& vec)
{
}


/* train multiple times with the same feature set 
and select the GMM parameters with the best classification 
performance. GMM is very sensitive to the initialization 
and this way better initialization can be achieved. 
*/

float 
GMM::evaluate(FeatMatrix& featMatrix)
{
  cout << "GMM::EVALUATE CALLED" << endl;
  int i=0;
  unsigned int classSize = featMatrix.labelSize(0);
  if (classSize != classSize_)
    {
      cerr << "FeatMatrix used for training and FeatMatrix for evaluation have different # of classes" << endl;
      return 1.0;
    }
  
  unsigned int r;
  unsigned int rows = featMatrix.rows();
  unsigned int cols = featMatrix.cols();
  unsigned int temp;
  unsigned int wrong =0;
  unsigned int maxwrong =0;
  unsigned int right = 0;
  unsigned int max = 0;
  
  for (i=0; i < eiterations_; i++)
    {
      wrong = 0;
      train(trainMatrix_,0);
      
      fvec vec(cols);
      for (r=0; r < rows; r++)
	{
	  vec = featMatrix.row(r);
	  
	  temp  = classify(vec);
	  if (featMatrix.labelNum(r,0) != temp)
	    {
	      wrong++;
	    }
	}
      right = rows - wrong;
      if (right > max) 
	{
	  max = right;
	  maxwrong = wrong;
	}
    }
  if (maxwrong == 0.0) return 0.0;
  else 
    return (maxwrong * 1.0/ rows * 1.0);
}



void 
GMM::initialize()
{
  unsigned int cl,k,r,c;
  unsigned int trainSize = trainMatrix_->rows();
  //unsigned int chunkSize = trainSize / (classSize_ * k_);
  
  fvec temp;
  temp.create(featSize_);
  fvec randstep(featSize_);
  temp.setval(0.0);
  unsigned int count;
  float rind = 0.0;
  rind = ((float)rand() / (float)(RAND_MAX))*trainSize;
  unsigned int seedSize = 5;
  for (cl=0; cl < classSize_; cl++)
    {
      for (k=0; k < k_; k++)
	{
	  temp.setval(0.0);
	  count = 0;
	  for (r=0; r < seedSize; r++)
	    {
	      rind = ((float)rand() / (float)(RAND_MAX))*trainSize;
	      while (trainMatrix_->labelNum((int)rind, 0) != cl) 
		rind = ((float)rand() / (float)(RAND_MAX))*trainSize;
	      //cout << "RIND = " << rind << endl;
	      temp += trainMatrix_->row((int)rind);
	    }
	  temp /= seedSize;
	  means_[cl].row(k, temp);
	  
	  temp = trainMatrix_->varRow();
	  vars_[cl].row(k, temp);
	}
      
      for (k=0; k < k_; k++)
	for (c=0; c < featSize_; c++)
	  {
	    if ((vars_[cl](k,c) != 0.0))
	      ((covars_[cl])(k,c)) = 1.0 / ((vars_[cl])(k,c));
	    else 
	      ((covars_[cl])(k,c)) = 0.0;
	  }
      weights_[cl].setval(1.0 / k_);

      // cout << "Means = " << means_[cl] << endl;
    }
  
  /* Perform k-means */ 
  
  
  float dist = 0.0;
  float min = 100000000;
  unsigned int min_k = 0;
  
  likelihoods_.create(classSize_, k_);
  unsigned int i;
  for (i=0; i < kiterations_; i++) 
    {
      likelihoods_.setval(0.0);
      for (cl = 0; cl < classSize_; cl++)
	for (k=0; k < k_; k++)
	  for (c=0; c < featSize_; c++)
	    {
	      omeans_[cl](k,c) = means_[cl](k,c);
	    }
      for (cl=0; cl < classSize_; cl++)
	{
	  means_[cl].setval(0.0);
	}
      for (r=0; r < trainSize; r++) 
	{
	  cl = trainMatrix_->labelNum(r,0);
	  temp = trainMatrix_->row(r);	  
	  min = 100000000;
	  for (k=0; k < k_; k++)
	    {
	      dist = mahalanobis(temp, (omeans_[cl]).row(k), (covars_[cl]).row(k));
	      // cout << "k cl dist = " << k << " " << cl << " " << dist << endl;
	      if (dist < min) 
		{
		  min = dist;
		  min_k = k;
		}
	    }
	  //cout << "MIN_K = " <<  min_k << " for " << cl << endl;
	  // cout << "KMEANS CLASS = " << cl << endl;
	  // cout << "Means = " << omeans_[cl] << endl;
	  
	  for (c=0; c < featSize_; c++)
	    {
	      means_[cl](min_k,c) += temp(c);
	    }
	  likelihoods_(cl,min_k)++;
	  
	}
      
      // cout << "Likelihoods = " << likelihoods_ << endl;
      
      /* for (cl=0; cl < classSize_; cl++)
	{
	  cout << "CLASS = " << cl << endl;
	  cout << "Means = " << means_[cl] << endl;
	}
      */
      for (cl=0; cl < classSize_; cl++)
	{
	  
	  for (k=0; k < k_; k++)
	    for (c=0; c < featSize_; c++)
	      {
		if (likelihoods_(cl,k) != 0.0)
		  means_[cl](k,c) /= likelihoods_(cl, k);
	      }

	  

	  // cout << "KMEANS CLASS = " << cl << endl;
	  // cout << "Means = " << means_[cl] << endl;
	}	
      
	  /*for (cl=0; cl < classSize_; cl++)
	{
	}
      */
      
    }

  likelihoods_.setval(0.0);      

  

  classSizes_.create(classSize_);
  sum_.create(classSize_);
  likelihoods_.create(classSize_, k_);
  probs_.reserve(classSize_);
  ssprobs_.reserve(classSize_);
  accumrow_.create(featSize_);
  row_.create(featSize_);
  temp_.create(featSize_);
  
  for (cl=0; cl < classSize_; cl++)
    {
      fmatrix clprob;
      fmatrix clssprob;
      clprob.create(trainSize, k_);
      clssprob.create(k_, featSize_);
      probs_.push_back(clprob);
      ssprobs_.push_back(clssprob);
    }
  sprobs_.create(classSize_,k_);
}

double
GMM::gaussian(unsigned int cl, unsigned int k, fvec& vec)
{
  unsigned int c;
  double res;
  double oldres;
  double temp;
  float det = 1.0;
  for (c=0; c < featSize_; c++)
    det *=  (vars_[cl])(k,c);
  res = 1 / (factor_ * det);
  oldres = res;
  temp = mahalanobis(vec, (means_[cl]).row(k), (covars_[cl]).row(k));
  res *= exp(-temp*0.5);
  return res;
}


void 
GMM::update()
{
  unsigned int r,c;
  unsigned int rows;
  unsigned int k;
  unsigned int cl;
  classSizes_.setval(0.0);
  sum_.setval(0.0);
  rows = trainMatrix_->rows();
  sprobs_.setval(0.0);
  float prob;
  accumrow_.setval(0.0);
  
  for (cl=0; cl < classSize_; cl++)
    ssprobs_[cl].setval(0.0);
  
  float sum;
  
  for (r=0; r < rows; r++)
    {
      cl = trainMatrix_->labelNum(r,0);
      classSizes_(cl)++;
      sum = 0.0;
      row_ = trainMatrix_->row(r);
      for (k=0; k < k_; k++)
	{
	  likelihoods_(cl,k) = gaussian(cl,k, row_);
	  sum += likelihoods_(cl,k);
	}
      for (k=0; k < k_; k++)
	{
	  if (sum != 0.0) 
	    prob = likelihoods_(cl,k) / sum;
	  else 
	    {
	      prob = 0.0000000001;
	    }
	  (probs_[cl])(r,k) = prob;
	  sprobs_(cl,k) += prob;
	  temp_ = row_;
	  temp_ *= prob;
	  accumrow_ = (ssprobs_[cl]).row(k);
	  accumrow_ += temp_;
	  (ssprobs_[cl]).row(k, accumrow_);
	}
    }
  
  for (cl = 0; cl < classSize_; cl++)
    for (k=0; k < k_; k++)
      {
	(weights_[cl])(k) = sprobs_(cl,k) / classSizes_(cl);
	temp_ = ssprobs_[cl].row(k);
	if (sprobs_(cl,k) != 0.0)
	  {
	    temp_ /= sprobs_(cl,k);
	    (means_[cl]).row(k,temp_);
	  }
      }
  
  for (cl=0; cl < classSize_; cl++)
    ssprobs_[cl].setval(0.0);

  for (r=0; r < rows; r++)
    {
      cl = trainMatrix_->labelNum(r,0);
      row_ = trainMatrix_->row(r);
      for (k=0; k < k_; k++)
	{
	  prob = (probs_[cl])(r,k);
	  temp_ = row_;
	  temp_ -= means_[cl].row(k);
	  temp_.sqr();
	  temp_ *= prob;
	  accumrow_ = (ssprobs_[cl]).row(k);
	  accumrow_ += temp_;
	  (ssprobs_[cl]).row(k, accumrow_);
	}
    }
  for (cl = 0; cl < classSize_; cl++)
    {
      for (k=0; k < k_; k++)
	{
	  temp_ = ssprobs_[cl].row(k);
	  temp_ *= (1.0 / (sprobs_(cl,k) -1.0));
	  (vars_[cl]).row(k, temp_);
	}
      for (k=0; k < k_; k++)
	for (c=0; c < featSize_; c++)
	  {
	    if ((vars_[cl](k,c) > 0.0))
	      ((covars_[cl])(k,c)) = 1.0 / ((vars_[cl])(k,c));
	    else 
	      {
		((covars_[cl])(k,c)) = 10000000;
		((vars_[cl])(k,c)) = 0.0000001;
	      }
	  }
    }
  return;  
}



void
GMM::train(FeatMatrix* featMatrix,unsigned int lcol)
{
  unsigned int cl;
  lcol_ = lcol;
  trainMatrix_ = featMatrix;
  classSize_ = trainMatrix_->labelSize(lcol);
  featSize_ = trainMatrix_->cols();
  factor_ = pow((float)SQRT_TWO_PI, (int)featSize_);

  classNames_ = trainMatrix_->labelNames();
  featNames_ = trainMatrix_->featNames();

  means_.reserve(classSize_);
  omeans_.reserve(classSize_);
  vars_.reserve(classSize_);
  covars_.reserve(classSize_);
  weights_.reserve(classSize_);


  
  for (cl=0; cl < classSize_; cl++)
    {
      fmatrix cmeans;
      fmatrix ocmeans;
      fmatrix cvars;
      fmatrix ccovars;
      fvec cweights;
      cmeans.create(k_, featSize_);
      ocmeans.create(k_, featSize_);
      cvars.create(k_, featSize_);
      ccovars.create(k_, featSize_);
      cweights.create(k_);
      
      // Vectors of fmatrices
      means_.push_back(cmeans);
      omeans_.push_back(ocmeans);
      vars_.push_back(cvars);
      covars_.push_back(ccovars);
      weights_.push_back(cweights);
    }
  initialize();
  
  cout << "GMM::update " << iterations_ << " iterations" << endl;
  unsigned int i;
  for (i=0; i < iterations_ ; i++)
    {
      for (cl=0; cl < classSize_; cl++)
	{
	  // cout << "CLASS = " << cl << endl;
	  // cout << "Means = " << means_[cl] << endl;
	  // cout << "Vars = " << vars_[cl] << endl;
	  // cout << "Weights = " << weights_[cl] << endl;
	}
      // cerr << "GMM::iteration " << i << endl;
      update();
    }
  cout << endl;
}



unsigned int 
GMM::classify(fvec& vec)
{
  unsigned int cl,k;
  float maxProb = 0.0;
  unsigned int maxClass = 0;
  float prob;
  for (cl=0; cl < classSize_; cl++)
    {
      for (k=0; k < k_; k++)
	{
	  likelihoods_(cl,k) = (weights_[cl])(k) / mahalanobis(vec, means_[cl].row(k), covars_[cl].row(k));
	}
      prob = 0.0;
      for (k=0; k < k_; k++)
	{
	  prob += (likelihoods_(cl,k));
	}
      if (prob > maxProb) 
	{
	  maxProb = prob;
	  maxClass = cl;
	}
    }
  return maxClass;
}



ostream&
operator<<(ostream& o, GMM& gmm)
{
  unsigned int i;
  o << "# MARSYAS Classifier " << endl;
  o << "# name = " << gmm.name_ << endl;
  o << "# type = " << gmm.type_ << endl << endl;
  o << "# classSize = " << gmm.classSize_ << endl;
  o << "# featSize = " << gmm.featSize_ << endl;
  o << "# components = " << gmm.k_ << endl;
  o << endl << endl;
  o << "# clasNames = " << endl;
  for (i=0; i< gmm.classSize_; i++)
    {
      o << gmm.classNames_[i] << endl;
    }
  o << endl;
  o << "# featNames = " << endl;
  for (i=0; i< gmm.featSize_; i++)
    {
      o << gmm.featNames_[i] << endl;
    }
  
  o << endl;

  unsigned int cl;
  
  for (cl=0; cl < gmm.classSize_; cl++)
    {
      o << "# Class " << cl << " = " << endl;
      o << "# Means = " << endl;
      o << gmm.means_[cl];
      o << endl;
      o << "# Vars = " << endl;
      o << gmm.vars_[cl];
      o << endl;
      o << "# Weights = " << endl;
      o << endl;
      o << gmm.weights_[cl];
    }
  return o;

}

void 
GMM::read(string filename)
{
  ifstream from(filename.c_str());
  from >> (*this);
}


void 
GMM::write(string filename)
{
  ofstream to(filename.c_str());
  to << (*this) << endl;
}


istream&
operator>>(istream& is, GMM& gmm)
{
  unsigned int cl,c,k;
  string str0, str1, str2;
  is >> str0 >> str1 >> str2;
  
  if ((str0 != "#") || ( str1 != "MARSYAS") || (str2 != "Classifier"))
    {
      cerr << "Problem reading classifier from istream" << endl;
      return is;
    }
  is >> str0 >> str1 >> str2;
  is >> gmm.name_;
  is >> str0 >> str1 >> str2;
  is >> gmm.type_;
  is >> str0 >> str1 >> str2;
  is >> gmm.classSize_;
  is >> str0 >> str1 >> str2;
  is >> gmm.featSize_;
  is >> str0 >> str1 >> str2;
  is >> gmm.k_;
  
  is >> str0 >> str1 >> str2;
  for (cl=0; cl < gmm.classSize_; cl++)
    {
      is >> str0;
      gmm.classNames_.push_back(str0);
    }
  is >> str0 >> str1 >> str2;
  for (c=0; c < gmm.featSize_; c++)
    {
      is >> str0;
      gmm.featNames_.push_back(str0);
    }
  gmm.means_.reserve(gmm.classSize_);
  gmm.vars_.reserve(gmm.classSize_);
  gmm.covars_.reserve(gmm.classSize_);
  gmm.weights_.reserve(gmm.classSize_);
  fmatrix cmeans;
  fmatrix cvars;
  fmatrix ccovars;
  for (cl=0; cl < gmm.classSize_; cl++)
    {
      fvec cweights;
      is >> str0 >> str1 >> str0 >> str1;
      is >> str0 >> str1 >> str0;
      is >> cmeans;
      is >> str0 >> str1 >> str0;
      is >> cvars;
      is >> str0 >> str1 >> str0;
      is >> cweights;
      ccovars.create(gmm.k_, gmm.featSize_);
      gmm.means_.push_back(cmeans);
      gmm.vars_.push_back(cvars);
      gmm.covars_.push_back(ccovars);
      gmm.weights_.push_back(cweights);
      for (k=0; k < gmm.k_; k++)
	for (c=0; c < gmm.featSize_; c++)
	  {
	    if ((gmm.vars_[cl](k,c) != 0.0))
	      ((gmm.covars_[cl])(k,c)) = 1.0 / ((gmm.vars_[cl])(k,c));
	    else 
	      ((gmm.covars_[cl])(k,c)) = 0.0;
	  }
    }
  
  gmm.classSizes_.create(gmm.classSize_);
  gmm.sum_.create(gmm.classSize_);
  gmm.likelihoods_.create(gmm.classSize_, gmm.k_);
  gmm.accumrow_.create(gmm.featSize_);
  gmm.row_.create(gmm.featSize_);
  gmm.temp_.create(gmm.featSize_);
  return is;  
  
}




