/*
** Copyright (C) 2000 George Tzanetakis <gtzan@cs.princeton.edu>
**  
** This program is free software; you can redistribute it and/or modify
** it under the terms of the GNU General Public License as published by
** the Free Software Foundation; either version 2 of the License, or
** (at your option) any later version.
** 
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU General Public License for more details.
** 
** You should have received a copy of the GNU General Public License
** along with this program; if not, write to the Free Software 
** Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
*/

/** 
    \class Gaussian.h
    \brief Gaussian classifier

    Gaussian classifier is a classifier where each class is modelled as 
a single multidimensional Gaussian distribution. The features 
are assumed to be uncorrelated and a diagonal covariance matrix 
is used.
*/


/* Gaussian classifier:
   Gaussian classifier.
*/


#include "Gaussian.h"

Gaussian::Gaussian()
{
  type_ = "Gaussian";
  name_ = "DefaultName";
  lcol_ = 0;
}

Gaussian::~Gaussian()
{
  
}




void
Gaussian::train(FeatMatrix* featMatrix, unsigned int lcol)
{
  unsigned int cl;
  unsigned int c;
  
  lcol_ = lcol;
  trainMatrix_ = featMatrix;
  classSize_ = trainMatrix_->labelSize(lcol);
  featSize_ = trainMatrix_->cols();

  classNames_ = trainMatrix_->labelNames();
  featNames_ = trainMatrix_->featNames();

  means_.create(classSize_, featSize_);
  vars_.create(classSize_, featSize_);
  covars_.create(classSize_, featSize_);
  trainMatrix_->classStat(lcol, means_, vars_);
  mask_.create(classSize_);
  mask_.setval(0);
  
  for (cl =0; cl < classSize_; cl++)
    for (c = 0; c < featSize_; c++)
      {
	if (vars_(cl,c) <= 0.000001)
	  covars_(cl,c) = 1.0;	// No variance
	else
	  {
	    covars_(cl,c) = 1.0 / vars_(cl,c);
	    mask_(cl) = 1;
	  }
      }
}



unsigned int 
Gaussian::classify(fvec& vec)
{
  float dist;
  int vec_class=0;
  unsigned int cl,c;
  float min = 100000.0;
  
  for (cl = 0; cl < classSize_; cl++)  
    {
      dist = mahalanobis(vec, means_.row(cl), covars_.row(cl));
      if ((dist < min)&&(mask_(cl)==1))
	{
	  min = dist;
	  vec_class = cl;
	}
    }

  return vec_class;
}


ostream&
operator<<(ostream& o, Gaussian& gaussian)
{
  unsigned int i;
  o << "# MARSYAS Classifier " << endl;
  o << "# name = " << gaussian.name_ << endl;
  o << "# type = " << gaussian.type_ << endl << endl;
  o << "# classSize = " << gaussian.classSize_ << endl;
  o << "# featSize = " << gaussian.featSize_ << endl;

  o << endl << endl;
  o << "# clasNames = " << endl;
  for (i=0; i< gaussian.classSize_; i++)
    {
      o << gaussian.classNames_[i] << endl;
    }
  o << endl;
  o << "# featNames = " << endl;
  for (i=0; i< gaussian.featSize_; i++)
    {
      o << gaussian.featNames_[i] << endl;
    }
  o << endl;


  o << endl << endl;
  o << "# Class Means = " << endl;
  o << gaussian.means_ << endl;
  o << "# Class Vars = " << endl;
  o <<  gaussian.vars_ << endl;
  return o;
}

void 
Gaussian::read(string filename)
{
  ifstream from(filename.c_str());
  from >> (*this);
}


void 
Gaussian::write(string filename)
{
  ofstream to(filename.c_str());
  to << (*this) << endl;
}


istream&
operator>>(istream& is, Gaussian& gaussian)
{
  unsigned int cl,c;
  string str0, str1, str2;
  is >> str0 >> str1 >> str2;
  
  if ((str0 != "#") || ( str1 != "MARSYAS") || (str2 != "Classifier"))
    {
      cerr << "Problem reading classifier from istream" << endl;
      return is;
    }
  is >> str0 >> str1 >> str2;
  is >> gaussian.name_;
  is >> str0 >> str1 >> str2;
  is >> gaussian.type_;
  is >> str0 >> str1 >> str2;
  is >> gaussian.classSize_;
  is >> str0 >> str1 >> str2;
  is >> gaussian.featSize_;
  
  is >> str0 >> str1 >> str2;
  for (cl=0; cl < gaussian.classSize_; cl++)
    {
      is >> str0;
      gaussian.classNames_.push_back(str0);
    }
  
  is >> str0 >> str1 >> str2;
  for (c=0; c < gaussian.featSize_; c++)
    {
      is >> str0;
      gaussian.featNames_.push_back(str0);
    }
  
  is >> str0 >> str1 >> str0;
  is >> str0;
  is >> gaussian.means_;
  is >> str0 >> str1 >> str0;
  is >> str0;
  is >> gaussian.vars_;
  gaussian.covars_.create(gaussian.classSize_, gaussian.featSize_);

  gaussian.mask_.create(gaussian.classSize_);
  gaussian.mask_.setval(0);

  for (cl =0; cl < gaussian.classSize_; cl++)
    for (c = 0; c < gaussian.featSize_; c++)
      {
	if (gaussian.vars_(cl, c) <= 0.0001)
	  gaussian.covars_(cl,c) = 1.0;		// No variance
	else
	  {
	    gaussian.covars_(cl,c) = 1.0 / gaussian.vars_(cl,c);
	    gaussian.mask_(cl) = 1;
	  }
      }
  return is;  
  
}

