/*
** Copyright (C) 2000 George Tzanetakis <gtzan@cs.princeton.edu>
**  
** This program is free software; you can redistribute it and/or modify
** it under the terms of the GNU General Public License as published by
** the Free Software Foundation; either version 2 of the License, or
** (at your option) any later version.
** 
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU General Public License for more details.
** 
** You should have received a copy of the GNU General Public License
** along with this program; if not, write to the Free Software 
** Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
*/

/**
   \class  GenreExtractor
   \brief  Single Vector feature extractor for genre information

   Combination of SVFFT (spectral timbral features), 
BEAT (beat related information calculated using the DWT). 
and MPITCH (pitch related information)
*/



#include "GenreExtractor.h"

GenreExtractor::GenreExtractor(MSignal* in):Extractor(in)
{
  unsigned int oldwinSize = in->winSize();
  unsigned int oldhopSize = in->hopSize();
  
  waveletSize_ = 65536;
  fftSize_ = 512;

  // Prepare BeatExtractor
  src_->initWindow(waveletSize_, waveletSize_/2,0,0);
  beatExtractor_ = new BeatExtractor(src_);
  
  // Prepare PitchExtractor 
  src_->initWindow(fftSize_, fftSize_,0,0);
  mpitchExtractor_ = new MultiPitchExtractor(src_);
  

  // Prepare TimbreExtractor 
  src_->initWindow(fftSize_, fftSize_,0,0);
  spectral_ = new Spectral(src_);
  mfcc_ = new MemMFCC(src_);
  svspectralExtractor_ = new SvFeatExtractor(src_, spectral_);
  svmfccExtractor_ = new SvFeatExtractor(src_, mfcc_);  
  
  unsigned int inSize = in->winSize();
  
  featSize_ = beatExtractor_->outSize() + mpitchExtractor_->outSize() + svspectralExtractor_->outSize() + svmfccExtractor_->outSize();
  
  // Change to get feature names from the extractors 

  unsigned  int i;
  for (i=0; i < featSize_; i++)
    {
      featNames_.push_back("gf");    
    }
  src_->initWindow(oldwinSize, oldhopSize,0,0);
}

GenreExtractor::~GenreExtractor()
{
  delete beatExtractor_;
  delete mpitchExtractor_;
  delete spectral_;
  delete mfcc_;
  delete svspectralExtractor_;
  delete svmfccExtractor_;
}

unsigned int 
GenreExtractor::outSize() 
{
  return featSize_;
}


void
GenreExtractor::extract()
{
}

void
GenreExtractor::extract(unsigned int start, unsigned int end, FeatMatrix& fmatrix)
{
  int load_matrix = 1;
  unsigned int t,i;
  
  // Check that fmatrix passed for loading has the correct dimensions 
  if ((fmatrix.rows() != 1) || (fmatrix.cols() != featSize_))
    {
      cerr << "GenreExtractor::extract incorrect matrix dimensions " << endl;
      cerr << "Rows = " << fmatrix.rows() << endl;
      cerr << "Cols = " << fmatrix.cols() << endl;
      cerr << "featSize_ = " << featSize_ << endl;
      load_matrix = 0;
    }
  unsigned int p=0;
  fmatrix.setFeatNames(featNames_);
  
  // Calculate beat features 
  FeatMatrix beatRes(1, beatExtractor_->outSize());
  src_->initWindow(waveletSize_, waveletSize_/2,0,0);  
  beatExtractor_->extract(0, src_->iterations(), beatRes);
  
  // Calculate timbre features 
  src_->initWindow(fftSize_, fftSize_,0,0);  
  FeatMatrix timbreRes1(1, svspectralExtractor_->outSize());
  FeatMatrix timbreRes2(1, svmfccExtractor_->outSize());
  svspectralExtractor_->extract(0, src_->iterations(), timbreRes1);
  svmfccExtractor_->extract(0, src_->iterations(), timbreRes2);
  
  // Calculate pitch features 
  FeatMatrix pitchRes(1, mpitchExtractor_->outSize());
  src_->initWindow(fftSize_, fftSize_,0,0);
  mpitchExtractor_->extract(0, src_->iterations(), pitchRes);

  // Load the full matrix 
  unsigned int tsize=beatExtractor_->outSize();
  for (i=0; i<tsize; i++) 
    fmatrix(0, i)=beatRes(0, i);

  tsize=svspectralExtractor_->outSize();
  unsigned int offset = beatExtractor_->outSize(); 
  for (i=0; i<tsize; i++)
    fmatrix(0, i+offset)=timbreRes1(0, i);

  tsize=svmfccExtractor_->outSize();
  offset = beatExtractor_->outSize() + svspectralExtractor_->outSize(); 
  for (i=0; i<tsize; i++)
    fmatrix(0, i+offset)=timbreRes2(0, i);
  
  
  tsize=mpitchExtractor_->outSize();
  offset=beatExtractor_->outSize()+svspectralExtractor_->outSize() + svmfccExtractor_->outSize();
  for (i=0; i<tsize; i++)
    fmatrix(0, i+offset)=pitchRes(0, i);   
  
  return;
}


void
GenreExtractor::extract(unsigned int start, unsigned int end)
{
  unsigned int t;
  for (t=start; t< end; t++)
    {
      if (t % ((end-start)/10))
	{
	  cerr << (end-start)/t * 100.0 << "% processed" << endl;
	}      
    }
}

