/*
** Copyright (C) 2000 George Tzanetakis <gtzan@cs.princeton.edu>
**  
** This program is free software; you can redistribute it and/or modify
** it under the terms of the GNU General Public License as published by
** the Free Software Foundation; either version 2 of the License, or
** (at your option) any later version.
** 
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU General Public License for more details.
** 
** You should have received a copy of the GNU General Public License
** along with this program; if not, write to the Free Software 
** Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
*/


/**
   \class  MultiPitchExtractor
   \brief  Single Vector feature extractor for beat information

   Feature vector of beat related information calculated using the DWT. 
*/



#include "MultiPitchExtractor.h"

MultiPitchExtractor::MultiPitchExtractor(MSignal* in):Extractor(in)
{
  int i;
  featSize_ = 5;
  for (i=0; i < featSize_; i++)
    featNames_.push_back("pf");

  in_.create(in->winSize());

  esacf_.create(in->winSize());
  multipitch_ = new MultiPitch(in->winSize());
  res_.create(featSize_);
  picker_ = new PeakPicker(in->winSize(), 1, 3, 22, in->winSize(), 50);
  peaks_.create(2 * 3);
  histo_.create(130);
  fhisto_.create(15);
}

MultiPitchExtractor::~MultiPitchExtractor()
{
  delete multipitch_;
  delete picker_;
}

unsigned int 
MultiPitchExtractor::outSize() 
{
  return featSize_;
}


void
MultiPitchExtractor::extract()
{
  extract(0, src_->iterations());
}

void 
MultiPitchExtractor::extract(unsigned int start, unsigned int end, 
		      FeatMatrix& featMatrix) 
{
  int load_matrix = 1;
  unsigned int t;

  if ((featMatrix.rows() != 1) || (featMatrix.cols() != featSize_))
    {
      cerr << "MultiPitchExtractor::extract incorrect matrix dimensions " << endl;
      cerr << "Rows = " << featMatrix.rows() << endl;
      cerr << "Cols = " << featMatrix.cols() << endl;
      load_matrix = 0;
    }
  
  featMatrix.setFeatNames(featNames_);
  
  
  unsigned int p=0;
  unsigned int i;
  
  for (t=start; t< end; t++)
    {
      src_->get(in_, t);  
      multipitch_->process(in_, esacf_);
      picker_->process(esacf_, peaks_);
      
      // convert peaks to histogram bins
      float pitch;
      unsigned int bin;
      unsigned int fbin;
      
      float amp;
      float mpos;
      
      for (i=0; i < peaks_.size()/2; i++)
	{
	  mpos = peaks_(i + peaks_.size()/2);
	  if (mpos > 0.0)
	    {
	      pitch=22050.0/mpos;	     
	      fbin = pitchfind(pitch);
	      bin = pitchfind_unfolded(pitch);
	      //            cout<<"pitchfind="<<pitch<<endl;
	      // bin = pitchfind_unfolded(pitch);
	      //amp = 1.0;
	      amp = peaks_(i);
	      assert(bin < histo_.size());
	      histo_(bin) += amp;
	      fhisto_(fbin) += amp;
	    }
	}
      if (t % ((end-start)/5) == 0)
	{
	  cerr << p * 20  << "% ";
	  p++;
	}
    }
  cerr << endl;
  cerr << "MultiPitchExtractor::extract completed" << endl;
  for (i=0; i < 12; i++)
    {
      histo_(i) = fhisto_(i);
    }
  
  
  // Calculate features 
  PitchHistoFeatures phf(histo_.size());
  phf.process(histo_, res_);
  featMatrix.row(0, res_);
}


void
MultiPitchExtractor::extract(unsigned int start, unsigned int end)
{
  unsigned int t;
  for (t=start; t< end; t++)
    {
      src_->get(in_, t);
      if (t % ((end-start)/10))
	{
	  cerr << (end-start)/t * 100.0 << "% processed" << endl;
	}      
    }
}

