
#include "Pitch.h"

Pitch::Pitch(string model)
{
  inSize_ = DEFAULT_WIN_SIZE;
	model_ = new PitchModel(model, inSize_);
  init();
}


Pitch::Pitch(string model, int inSize)
{
  inSize_ = inSize;
	model_ = new PitchModel(model, inSize_);
  init();
}


Pitch::~Pitch()
{
  if(train_) model_->writeModel();
  delete model_;
	delete envelope_;
}


void
Pitch::init()
{
  featNames_.push_back("pitch");
  featNames_.push_back("mag");

  outSize_ = 2;
	minBin_ = (int)(100 * inSize_ / 44100);
	maxBin_ = inSize_ / 2 - 1;
  maxpartials_ = 20;
	order_ = 40;
	
  fft_.create(inSize_);
  spectrum_.create(inSize_/2);
  quality_.create(inSize_/2);
  likelihood_.create(inSize_/2);
  partials_.create(inSize_/2);
	freqs_.create(inSize_/2);
	amps_.create(inSize_/2);
	hist_.create(inSize_/2);
	thist_.create(inSize_/2, maxpartials_+1);
	probs_.create(maxpartials_+1);
	indexes_.create(maxpartials_+1);
	parnum_.create(inSize_/2);
	ccfs_.create(order_);
	pars_.create(inSize_/2);	
  train_ = false;
	
  envelope_ = new DFCC(maxpartials_ * 2, order_);
}


inline float
snr(fvec& partials_, fvec& indexes_, fvec& spectrum_, PitchModel* model_)
{
    int i, p;
	  float pow=0, sum=0, tmp;
		for(i = 0; i < 20; i++) if(indexes_(i) != -1)
		{
		  p = (int)partials_((int)indexes_(i));
			while(spectrum_(p-1) > spectrum_(p)) p--;
			while(spectrum_(p+1) > spectrum_(p)) p++;
		  pow += spectrum_(p);
			spectrum_(p) = 0;
			tmp = spectrum_(p) / model_->hamfft_(2);
			if(spectrum_(p-1) > model_->hamfft_(1) * tmp)
			{
			  pow += model_->hamfft_(1) * tmp;
				spectrum_(p-1) -= model_->hamfft_(1) * tmp;
			}
			else
			{
			  pow += spectrum_(p-1);
				spectrum_(p-1) = 0;
			}
			if(spectrum_(p+1) > model_->hamfft_(3) * tmp)
			{
			  pow += model_->hamfft_(1) * tmp;
				spectrum_(p+1) -= model_->hamfft_(3) * tmp;
			}
			else
			{
			  pow += spectrum_(p+1);
				spectrum_(p+1) = 0;
	    }
		}
		for(i = 0; i < spectrum_.size(); i++) sum += spectrum_(i);
		if(sum > 0) return pow/sum;
		return 0;
}



inline float
snr(int npartials_, fvec& partials_, fvec& spectrum_, PitchModel* model_)
{
    int i, j, p;
	  float pow=0, sum=0, tmp;
		for(i = 0, j = 0; i < npartials_; i++)
		{
		  p = (int)partials_(i);
		  while(j < p-1) sum += spectrum_(j++)*spectrum_(j++);
		  pow += spectrum_(p)*spectrum_(p) + spectrum_(p-1)*spectrum_(p-1) + spectrum_(p+1)*spectrum_(p+1);
			j = p+2;
		}
		while(j < spectrum_.size()) sum += spectrum_(j++)*spectrum_(j++);
		if(sum > 0) return pow/sum;
		return 0;
}



void 
Pitch::process(fvec& in, fvec& out) 
{
  assert((in.size() == inSize_) && (out.size() == outSize_));  
	
  int i, j, k, p, candidate, min, max, num, finalists = 5;
	float s, cutoff;
	fvec highest(finalists);
	
  //calculate power spectrum, get qualities, pick partials, and create histogram
	npartials_ = model_->getPartials(in, spectrum_, fft_, quality_, partials_);
	for(i = 0; i < npartials_; i++)
	{
	  interpolate(partials_(i), out);
	  freqs_(i) = out(0);
		amps_(i) = out(1);
	}
	
	out.setval(0);
	if(npartials_ == 0) return; 
	if(snr(npartials_, partials_, spectrum_, model_) < 5) return;
  //cout << snr(npartials_, partials_, spectrum_, model_) << endl;
 
	for(i = 0; i < npartials_; i++)
	{
		pars_(i) = 2 * PI * freqs_(i) / inSize_;
		pars_(i+npartials_) = amps_(i);
	}
	envelope_->set(npartials_*2, order_);
	envelope_->process(pars_, ccfs_);
 
	for(i = 0; i < npartials_; i++)
	{ 
	  quality_(i) = envelope_->get(ccfs_, 2*PI*freqs_(i)/inSize_);
		quality_(i) = model_->normpdf(amps_(i), quality_(i)/4, quality_(i));
	}
 
	hist_.setval(1);
	highest.setval(0);
	//match partials to harmonics, multiply likelihood by frequency probablity
	for(i = 0; i < (int)inSize_/2; i++)
	{
		p = match(i);
		for(j = 1; j <= maxpartials_; j++)
		{
		  if(indexes_(j) != -1) hist_(i) *= model_->prob(i, j) * probs_(j);
		  else hist_(i) *= 1 - model_->prob(i, j);
		}
	  for(j = finalists; j > 0 && hist_(i) > highest(j-1); j--);
		if(j == 0 || (j != finalists && hist_(i) != highest(j-1)))
		{
	    for(k = finalists-1; k > j; k--) highest(k) = highest(k-1);
			highest(k) = hist_(i);
		}
	}
	for(j = 0; j < finalists-1 && highest(j+1) != 0; j++);
	cutoff = highest(j);

  //calculate liklihoods for selected candidates
	likelihood_.setval(0);
	for(candidate = minBin_; candidate < maxBin_; candidate++) if(hist_(candidate) >= cutoff)
	{
		p = match(candidate);
		likelihood_(candidate) = hist_(candidate);
		
		//get dfcc's and multiply liklihood by un-matched partials' amplitude probability
	  for(i = 1, j = 0; i <= maxpartials_ && j < p; i++) if(indexes_(i) != -1)
		{
		  pars_(j) = 2 * PI * freqs_((int)indexes_(i)) / inSize_;
		  pars_(j+p) = amps_((int)indexes_(i));
			j++;
		}
	  envelope_->set(p*2, order_);
	  envelope_->process(pars_, ccfs_);
	  for(i = 0; i < npartials_; i++) if(parnum_(i) <= 0)
	  {
		  float s=amps_(i), a=envelope_->get(ccfs_, 2*PI*freqs_(i)/inSize_);
		  likelihood_(candidate) *= model_->normpdf(s,a/4,0)/model_->normpdf(0,a/4,0);
	  }

		//choose candidate with the highest likelihood
		if(likelihood_(candidate) > out(1))
		{
			out(0) = candidate;
		  out(1) = likelihood_(candidate);
		}
  }
	
	if(train_) model_->add(partials_, indexes_, out(0));


	for(i = 0; i < npartials_; i++)
	{
		pars_(i) = 2 * PI * freqs_(i) / inSize_;
		pars_(i+npartials_) = amps_(i);
	}
	cout << pars_;
	envelope_->set(npartials_*2, order_);
	envelope_->process(pars_, ccfs_);
		
		
	  ofstream hist("hist"), like("like"), spect("spectrum"), env("envelope");
	  for(i = 0; i < inSize_/2; i++)
	  {
	    hist << hist_(i) << endl;
	    like << likelihood_(i) << endl;
		  spect << spectrum_(i) << endl;
			env << envelope_->get(ccfs_, 2*PI*i/inSize_) << endl;
		}
	
	

	if(out(0) != 0)
	{
	  if(spectrum_(out(0)) > spectrum_(out(0)+1) && spectrum_(out(0)) > spectrum_(out(0)-1))
		  out(0) = interpolate(out(0), out);
		else if(spectrum_(out(0)-1) > spectrum_(out(0)) && spectrum_(out(0)-1) > spectrum_(out(0)-2))
		  out(0) = interpolate(out(0)-1, out);
		else if(spectrum_(out(0)+1) > spectrum_(out(0)) && spectrum_(out(0+1)) > spectrum_(out(0)+2))
		  out(0) = interpolate(out(0)+1, out);
		else out(0) = inst(out(0));
	}
}

void
Pitch::train()
{
  train_ = true;
}

inline int
Pitch::match(int candidate)
{
	int i, p, dist, forder, offset, num = model_->getModel(candidate);
	float prob, amp;

	probs_.setval(0);
	indexes_.setval(-1);
	parnum_.setval(0);
	
	for(i = 0; i < npartials_; i++)
	{
		p = round((float)freqs_(i)/candidate);
		if(p > 0 && p <= maxpartials_)
		{ 
		  offset = freqs_(i) - p * candidate;
			prob = model_->fdist(num, p, offset);
			if(indexes_(p) == -1 || prob*quality_(i) > probs_(p)*quality_(indexes_(p)))
			{
			  probs_(p) = prob;
				if(indexes_(p) != -1) parnum_((int)indexes_(p)) = -1;
				indexes_(p) = i;
				parnum_(i) = p;
			}
			else parnum_(i) = -1;
		}
	}
	for(i = 0, p = 0; i < maxpartials_; i++) if(indexes_(i) != -1) p++;
  return p;
}


inline int
Pitch::round(float r)
{
  if(r - floor(r) <= 0.5 ) return (int)floor(r);
  return (int)ceil(r);
}


//parabolic interpolation
inline float
Pitch::interpolate(int p, fvec& vals)
{
  int i = p;
  float a, g, b, P;
	if(spectrum_(p) <= spectrum_(p+1) || spectrum_(p) <= spectrum_(p-1))
	{
	  vals(0) = p;
		vals(1) = spectrum_(p);
		return 44100.0*p/inSize_;
	}
	a = 20*log10(spectrum_(i-1));
	b = 20*log10(spectrum_(i));
	g = 20*log10(spectrum_(i+1));
	P = (a-g)/(2 * (a-2*b+g));
	vals(0) = i+P;
	vals(1) = spectrum_(i)-(spectrum_(i-1)-spectrum_(i+1))*P/4;
	return 44100 * vals(0) / inSize_;
}


//instantaneous frequency
inline float
Pitch::inst(int p)
{
  float a, b, c, d, w = 2*PI/inSize_, cw = cos(w), sw = sin(w), ncw = cos(-w), nsw = sin(-w);
	float T, Ti, f;
	a = fft_(p*2)   - (fft_(p*2-2) + fft_(p*2+2)) / 2;
	b = fft_(p*2+1) - (fft_(p*2-1) + fft_(p*2+3)) / 2;
  c = fft_(p*2)   - (cw*fft_(p*2-2) - sw*fft_(p*2-1) + ncw*fft_(p*2+2) - nsw*fft_(p*2+3)) / 2;
  d = fft_(p*2+1) - (cw*fft_(p*2-1) + sw*fft_(p*2-2) + ncw*fft_(p*2+3) + nsw*fft_(p*2+2)) / 2;
	T = (a*c+b*d)/(c*c+d*d);
	Ti = (b*c-a*d)/(c*c+d*d);
	f = 44100*atan(Ti/T)/(2*PI);
	if(f < 0) return inst(p-1);
	return 44100*p/inSize_ + f;
}


