#include "PitchModel.h"


PitchModel::PitchModel(int inSize)
{
  inSize_ = inSize;
	name_ = "standard";
  init();
}


PitchModel::PitchModel(string model, int inSize)
{
  inSize_ = inSize;
  name_ = model;
  init();
}


PitchModel::~PitchModel()
{
}


void
PitchModel::init()
{
  magfft_ = new MagFFT(inSize_);
  hamming_ = new Hamming(inSize_, 0);
  
	maxpartials_ = 20;
	
  in_.create(inSize_);
  windowed_.create(inSize_);
  hamfft_.create(inSize_);
  spectrum_.create(inSize_/2);
  quality_.create(inSize_/2);
  partials_.create(inSize_/2);
	probs_.create(inSize_/2, maxpartials_+1);
	means_.create(inSize_/2, maxpartials_+1);
	vars_.create(inSize_/2, maxpartials_+1);
	norms_.create(inSize_/2, maxpartials_+1);
	stats_.create(inSize_/2, 3*maxpartials_+1);
	
	int i;
  for(i = 0; i < inSize_; i++) hamfft_(i) = sin(4*PI*i/inSize_);
  hamming_->process(hamfft_, windowed_);
  magfft_->process(windowed_, hamfft_);
  hamfft_ /= sqrt(pow(hamfft_(1), 2) + pow(hamfft_(2), 2) + pow(hamfft_(3), 2));
	
	filename_ = MARSYAS_MFDIR;
  filename_ += "/models/";
  filename_ += name_;
  filename_ += ".mdl";
	
	means_.setval(0);
	vars_.setval(0);
	norms_.setval(0);
	stats_.setval(0);
	
	createStandard();
  struct stat stbuf;
  if(stat(filename_.c_str(), &stbuf) || name_ == "standard") cout << "Using standard model" << endl;
  else readModel();

}


//returns number of partials
int 

PitchModel::getPartials(fvec& in, fvec& spectrum, fvec& fft, fvec& quality, fvec& partials)
{
  int i, npartials = 0;
	float sum, cutoff, q;
  hamming_->process(in, windowed_);
  magfft_->process(windowed_, spectrum); 
  for(i = 0, sum = 0; i < inSize_/2; i++) sum += pow(spectrum(i), 2);
  if(sqrt(sum)) spectrum /= sqrt(sum);
	fft = magfft_->fft();
  //calculate quality for each bin
  int minBin = (int)(100.0*inSize_/44100.0), maxBin = inSize_/2-1;
  for(i = minBin; i < maxBin; i++)
	  quality(i) = hamfft_(1)*spectrum(i-1)+hamfft_(2)*spectrum(i)+hamfft_(3)*spectrum(i+1);
	
	//return partials above cutoff value
  cutoff = .05;
  for(i = minBin; i < maxBin; i++)
  {
		q = quality(i);
    if(q > cutoff && q > quality(i-1) && q > quality(i+1)) partials(npartials++) = i;
  }
	
  return npartials;
}


float
PitchModel::fdist(int cand, int partial, float offset)
{
  return normpdf(offset, vars_(cand, partial), means_(cand, partial)) / norms_(cand, partial);
}


float
PitchModel::prob(int candidate, int partial)
{
  return probs_(candidate, partial);
}


void 
PitchModel::add(fvec& in, int label) 
{
	int i, p, pindex, npartials_;
	fvec fft;
	npartials_ = getPartials(in, spectrum_, fft, quality_, partials_);
  for(i = 0; i < npartials_; i++)
  {
    pindex = round(partials_(i));
    p = round((float)pindex/label);
    if(p > 0 && p <= maxpartials_)
    {
      stats_(label, p)++;
      stats_(label, p + maxpartials_) += pindex;
      stats_(label, p + 2*maxpartials_) += pindex*pindex;
    }
  }
  stats_(label, 0)++;
}


void 
PitchModel::add(fvec& pars, fvec& indexes, int label) 
{
	int i;
	float offset;
  for(i = 1; i <= maxpartials_; i++) if(indexes(i) != -1)
  {
    offset = partials_(indexes(i)) - i*label;
		stats_(label, i)++;
		stats_(label, i + maxpartials_) += offset;
		stats_(label, i + 2*maxpartials_) += offset*offset;
  }
  stats_(label, 0)++;
}


void 
PitchModel::addSource(MSignal *src, int label) 
{
  assert(src->winSize() == winSize_);
  int frame, samples = src->iterations();
  fvec in;
  in.create(inSize_);
  for(frame = 0; frame < samples; frame++)
  {
    cout << "frame " << frame << endl;
    src->get(in, frame);
    add(in, label);
  }
}


float
PitchModel::normpdf(float value, float var, float mean)
{
  return exp(-(value-mean)*(value-mean)/(2*var))/(sqrt(var)*sqrt(2*PI));
}


void
PitchModel::readModel()
{
	cout << "Reading model from " << filename_ << endl;
	ifstream mis(filename_.c_str());
	mis >> stats_;
	mis.close();
	updateModel();
	cout << "Model creation complete." << endl;	
}


void
PitchModel::writeModel()
{
	cout << "Writing model to " << filename_ << endl;
	mos.open(filename_.c_str());
	mos << stats_;
	mos.close(); 
}


void
PitchModel::createStandard()
{
  int i, j, m = maxpartials_ * maxpartials_;
	for(i = 0; i < inSize_/2; i++)
	{
		stats_(i, 0) = maxpartials_+1;
		for(j = 1; j <=maxpartials_; j++)
		{
			stats_(i, j) = stats_(i, 0) * normpdf(j, m, 0) / normpdf(0, m, 0);
			stats_(i, j + maxpartials_) = 0;
			stats_(i, j + 2 * maxpartials_) = pow(i*j/36.0, 2)*stats_(i, j);
		}
	}
	updateModel();
	stats_.setval(0);
}


int
PitchModel::getModel(int candidate)
{
  int top = candidate, bot = candidate;
  if(probs_(candidate, 0) > 0) return candidate;
  while(top < inSize_/2 && probs_(top, 0) == 0) top++;
  while(bot > 0 && probs_(bot, 0) == 0) bot--;
  if(bot == 0) return top;
  if(top == 0) return bot;
  if(candidate-bot < top-candidate) return bot;
  return top;
}


void
PitchModel::updateModel()
{
  int i, j;
	float prob;
	for(i = 0; i < inSize_/2; i++)
	 for(j = 1; j <= maxpartials_; j++)
	   if(probs_(i, 0) = stats_(i, 0))
	   {
		   prob = stats_(i, j);
		   probs_(i, j) = prob / stats_(i, 0);
		   if(prob)
		   {
			   means_(i, j) = stats_(i, j+maxpartials_)   / prob;
			   vars_(i, j)  = stats_(i, j+2*maxpartials_) / prob; 
			   vars_(i, j) -= means_(i, j) * means_(i, j);
		     norms_(i, j) = normpdf(0, vars_(i, j), 0);
		   }
	   }
}


inline int
PitchModel::round(float r)
{
  if(r - floor(r) <= 0.5 ) return (int)floor(r);
  return (int)ceil(r);
}



