
/*
** Copyright (C) 2000 George Tzanetakis <gtzan@cs.princeton.edu>
**  
** This program is free software; you can redistribute it and/or modify
** it under the terms of the GNU General Public License as published by
** the Free Software Foundation; either version 2 of the License, or
** (at your option) any later version.
** 
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU General Public License for more details.
** 
** You should have received a copy of the GNU General Public License
** along with this program; if not, write to the Free Software 
** Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
*/


/**
   \class Pitches
   \brief FFT based Pitches features 


   Calculate FFT mangitude based features. 
The Centroid, Rolloff and Flux are calculated 
on the magnitude spectrum of the input fvec 
and Zcrs are calculated in time domain. 
This System feature are typically used in classification. 
*/


#include "Pitches.h"

Pitches::Pitches()
{
  inSize_ = DEFAULT_WIN_SIZE;
  pitches_ = 1;
  outSize_ = 2 * pitches_; 
  init();
}

Pitches::Pitches(int inSize, int pitches)
{
  inSize_ = inSize;
  pitches_ = pitches;
  outSize_ = pitches*2; 
  init();
}


void 
Pitches::init()
{
  featSize_ = outSize_;
  int n, i;
  for(n = 0; n < pitches_; n++)
  {
    featNames_.push_back("pitch");
    featNames_.push_back("mag");
  }
  current_pitch = 0;
  esacf_.create(inSize_/2);
  hamming_ = new Hamming(inSize_, zeroSize_);
  magfft_  = new MagFFT(inSize_);  peaks_.create(outSize_);
  vector<System *> featArray;
  featArray.push_back(hamming_);
  featArray.push_back(magfft_);
  series_ = new Series(featArray);
  
  bins = 18; 
  rate = 44100.0;
  size = (float) inSize_;
  start = index(50.0);
  
  binStart = new int[bins];
  binSz = new int[bins];
  Xe = new (float *)[bins];
  Lb = new (float *)[bins];
  weights = new float[bins];
  candidates = new int[bins];
  bands = new int[bins];
  curves = new (int **)[bins];

  float f = 50.0, crt2 = pow(2, (float)1/3);
  for(i = 0; i < bins; i++)
  {
    binStart[i] = index(f);  
    cout << i << " " << f << " " << binStart[i];
    if(ceil(f * (crt2 * crt2 - 1)) >= 100)
    {
      f *= crt2;
      binSz[i] = index(f * crt2) - binStart[i];
    }
    else
    {
      binSz[i] = index(f + 100) - binStart[i];
      f += 40;
    }  
    cout << " " << binSz[i] << endl;
  }
  
  N = binStart[bins - 1] + binSz[bins - 1];
  for(i = 0; i < bins; i++)
  {
    Xe[i] = new float[binSz[i]];
    Lb[i] = new float[N];
    curves[i] = new (int *)[bins];
    for(int j = 0; j < bins; j++) curves[i][j] = new int[4];
  }
  ffts.open("ffts");
  offset = 0;
  lbs.open("lbs");
  lboffset = 0;
}



Pitches::~Pitches()
{
  delete hamming_;
  delete magfft_;
  delete series_;
  int j;
  for(j = 0; j < bins; j++)
  {
    delete Xe[j];
    delete Lb[j];
  }
  delete Lb;
  delete Xe;
  delete binStart;
  delete binSz;
}





void 
Pitches::process(fvec& in, fvec& out) 
{
  int i, n, Kb, kb;
  float sum = 0, mult;

  /*calculate power spectrum, filter out the noise, break into bands, and triangulate*/
  if(current_pitch == 0)
  {
    assert((in.size() == inSize_) && (out.size() == outSize_));    
    series_->process(in, esacf_);  
  }
  for(i = start; i < N; i++) sum += pow(esacf_(i), 1/3);
  sum = pow(sum / (N - start + 1), 3);
  for(i = start; i < N; i++) esacf_(i) = log(1 + esacf_(i)/sum);
  for(i = 0; i < bins; i++)
  {
    kb = binStart[i];
    Kb = binSz[i];
    for(n = 0, sum = 0; n < Kb; n++) sum += esacf_(kb + n);
    sum /= Kb;
    for(n = 0; n < Kb; n++) if((Xe[i][n] = esacf_(kb + n) - sum) <= 0) Xe[i][n] = 0;
    if(i == bins - 1) mult = binStart[i-1] + binSz[i-1] - kb;
    else mult = binStart[i+1] - kb;
    for(n = 0; n < Kb; n++) Xe[i][n] *= triangulate(n, Kb, mult);
  }

  /*calculate the bandwise likelihood vectors*/
  int lb, m, m0, m1, k, k0, k1, J, j;
  for(i = 0; i < bins; i++)
  {
    for(n = 0; n < N; n++) Lb[i][n] = 0;
    kb = binStart[i];
    Kb = binSz[i];
    lb = Kb + kb;
    for(n = start; n < Kb; n++)
    {
      m0 = round(ceil((float)kb/n) * n) - kb;
      m1 = m0 + (int)((lb - 1) * rate * (sqrt(1 + 0.01 * (pow((float)(lb-1)/n, 2) - 1.0)) - 1) / size);
      if(m1 > m0 + n)
      {
        m0 = 0;
        m1 = n;
      }
      for(m = m0; m <= m1; m++)
      {
        sum = 0.0;
        if((J = (int)floor((float)(Kb - m - 1) / n) + 1) > 1) for(j = 0; j < J; j++) sum += Xe[i][m + (n * j)];
        sum *= (.75/J + .25);
        if(sum > Lb[i][n]) Lb[i][n] = sum;
      }
    }
    j = 1;
    k0 = (int)floor((float)lb / 2);
    if(k0 < kb) k0 = kb;
    k1 = lb - 1;
    while(k0 <= k1)
    {//cout << " " << freq(round((float)k0/j)) << "-" << freq(round((float)k1/j)) << endl;
      for(k = k0; k <= k1; k++)
      {
        n = round((float)k/j);
        if(Lb[i][n] < Xe[i][k - kb]) Lb[i][n] = Xe[i][k - kb];
      }
      j++;
      k0 = (int)ceil((float)lb * j / (j + 1));
      if(k0 < kb) k0 = kb;
      k1 = (int)floor((float)(kb - 1) * j / (j - 1));
      if(k1 > lb) k1 = lb;
    }//cout << endl;
  }

  float current;
  int band;
  for(i = 0; i < bins; i++)
  {
    candidates[i] = 0;
    weights[i] = 0;
  }
  for(m = start; m < N; m++)
  {
    band = 0;
    for(i = 1; i < bins; i++) if(Lb[i][m] > Lb[band][m]) band = i;
    current = Lb[band][m];
    if(current > weights[bins-1])
    {
      for(i = bins - 1; i > 0 && candidates[i-1] == 0; i--);
      for(; i > 0 && current > weights[i-1]; i--)
      {
        weights[i]  = weights[i-1];
        candidates[i] = candidates[i-1];
        bands[i] = bands[i-1];
      }
      candidates[i]  = m;
      weights[i] = Lb[band][m];
      bands[i] = band;
    }
  }
  
  int candidate;
  for(i = 0; i < bins; i++) weights[i] = 0;
  for(i = 0; i < bins; i++) 
  {
    curve = curves[bins-1];
    curves[bins-1] = curves[i];
    candidate = candidates[bins-1];
    candidates[bins-1] = candidates[i];
    band = bands[bins-1];
    bands[bins-1] = bands[i];
    for(n = 0; n < bins; n++) curve[n][0] = 0;
    current = estimate(candidate, band, 0, 0.01);
    for(n = i; n > 0 && current > weights[n-1]; n--)
    {
      weights[n]  = weights[n-1];
      candidates[n] = candidates[n-1];
      curves[n] = curves[n-1];
    }
    candidates[n]  = candidate;
    weights[n] = current;
    curves[n] = curve;
  }

  int partials[50], partial, freq, gap, pfreq, I, d, p, max_candidate = 0;
  float *smooth, *max_smooth, max_likelihood = 0, weight = 0, total, likelihood;
  for(d = 0; d < bins && weight < weights[d]; d++)
  {
    smooth = new float[N];
    curve = curves[d];
    for(i = 0; i < 50; i++) partials[i] = 0;
    for(i = 0; i < N; i++) smooth[i] = 0;
    for(i = 0; i < bins && curve[i][0] == 0; i++) I = i;
    I = i;
    for(i = I; i < bins && curve[i][0] != 0; i++)
      for(n=0, partial=curve[i][0], freq=curve[i][1]; n<curve[i][3] && partial<50; n++, partial++, freq+=curve[i][2])
        if(partials[partial] == 0 || esacf_(freq) > esacf_(partials[partial])) partials[partial] = freq;
    while(partial < 50) { partials[partial] = partials[partial-1] + curve[i-1][2]; partial++; }
    for(i = I, likelihood = 0; i < bins && curve[i][0] != 0; i++)
    {
      if(curve[i][3] == 0) i++;
      kb = binStart[i];
      Kb = binSz[i];
      gap = curve[i][2];
      partial = curve[i][0];
      freq = curve[i][1];
      for(n = 0; n < curve[i][3]; n++, partial++, freq += gap) if(smooth[freq] == 0)
      {
        m0 = (int)floor(freq/sqrt(2));
        m1 = (int)ceil(freq*sqrt(2));
        p = (int)floor(partial/sqrt(2));
        while(partials[p] < m0) p++;
        for(sum = 0, total = 0; partials[p] < m1 && p < (int)ceil(partial * sqrt(2)) && p < 50; p++)
        { 
          if(p >= curve[i][0] && p < curve[i][0] + curve[i][3]) pfreq = freq + (p - partial) * gap;
          else pfreq = partials[p];
          weight = triangulate(pfreq - m0, m1 - m0, freq);
          total += weight;
          sum += esacf_(pfreq) * weight;
        }
        smooth[freq] = sum / total;
      }
      freq = curve[i][1];
      for(n = 0, sum = 0; n < Kb; n++)
      {
        if((kb + n - freq) % gap == 0) sum += smooth[kb + n];
        else sum += esacf_(kb + n);
      }
      sum /= Kb;
      if(i == bins - 1) mult = binStart[i-1] + binSz[i-1] - kb;
      else mult = binStart[i+1] - kb;
      for(n = freq - kb; n < Kb; n += gap)
        if(smooth[n + kb] - sum > 0) likelihood += (smooth[n + kb] - sum) * triangulate(n, Kb, mult);
    }
    if(likelihood > max_likelihood || d == 0)
    {
      if(d > 0) delete max_smooth;
      max_smooth = smooth;
      max_likelihood = likelihood;
      max_candidate = d;
    }
    else delete smooth;
    weights[d] = likelihood;
  }
  
  i=0;
  while(i < current_pitch && out(2*i) != candidates[max_candidate]) i++;
  if(i < current_pitch)
  {
    current_pitch = 0;
    cout << endl;
    return;
  }

  out(current_pitch * 2) = candidates[max_candidate];
  out(current_pitch * 2 + 1) = max_likelihood;

  current_pitch++;
  if(current_pitch < pitches_) for(i = 0; i < N; i++) if(esacf_(i)-max_smooth[i] >= 0) esacf_(i)-=max_smooth[i];

  printFfts();
  printWeights();
  ofstream smoothed("smoothed");
  for(i=0; i<N; i++) smoothed << i << " " << max_smooth[i] << endl;

  delete max_smooth;
  
  cout << candidates[max_candidate] << " " << max_likelihood << endl;
  if(current_pitch < pitches_) process(in, out);
  else
  {
    current_pitch = 0;
    cout << endl;
  }
  //for(i = 0; i < bins; i++) cout << candidates[i] << " " << weights[i] << endl;
  //cout << endl; 
  //for(i = 0; i < bins; i++) cout << curve[i][0]<< " "<<curve[i][1]<<" " << curve[i][2] << " " << curve[i][3] <<endl;


}


float
Pitches::estimate(int n, int band, float weight, float inharm)
{
    float sum = 0, inharm_next;
    int b, kb, Kb, lb, partial, freq, gap, maxgap, j, J, m, tmp[4];
    kb = binStart[band];
    Kb = binSz[band];
    lb = kb + Kb;
    
    /*base case: if fundamental freqency is not in this band, calculate the max inharmonicity for the next band*/
    if(weight == 0)
    {
      maxsum = 0;
      partial = (int)ceil((float)kb/n);
      //cout << "base: " << n << " " << band << endl;
      for(b = band; b >= 0; b--)
      {
        kb = binStart[b];
        Kb = binSz[b];
        lb = kb + Kb;
        if(lb > n)
        {
          curve[b][0] = (int)ceil((float)binStart[b]/n);
          curve[b][1] = n * curve[b][0];
          curve[b][2] = n;
          curve[b][3] = 1 + (int)floor((lb - curve[b][1] - 1) / n);
          sum += Lb[b][n];
        }
      }
      while(b >= 0) curve[b--][0] = 0;
      if(partial <= 6) inharm_next = inharm;
      else inharm_next = (pow(1 + 1.0/partial, 2) - 1) / (pow(partial, 2) - 1);
      if(band+1 == bins) return sum;
      return estimate(n, band + 1, sum, inharm_next);
    }
    //cout << band << endl;
    
    /*calculate the first partial in this band.  if there is no partial in this band go to the next*/
    partial = curve[band - 1][0];
    gap = curve[band - 1][2];
    freq = curve[band - 1][1];
    while(freq < kb)
    {
      freq += gap;
      partial++;
    }
    if(partial == 1 || freq >= lb)
    {
      curve[band][0] = 1;
      curve[band][1] = n;
      curve[band][2] = n;
      if(freq >= lb) curve[band][3] = 0;
      else curve[band][3] = 1;
      if(band+1 == bins) return weight + Lb[band][n];
      return estimate(n, band + 1, weight, inharm);      
    }
    
    /*test for maximum weight from all possible distances between adjecent partials*/
    float max = 0, prev, cur, next;
    prev = Lb[band][gap - 1];
    cur = Lb[band][gap];
    next = Lb[band][gap + 1];
    maxgap = (int)ceil(n * sqrt(1 + inharm * (pow(partial, 2) - 1)));
    while(gap <= maxgap)
    {
      if((2 * cur > prev + next && cur > 0 || cur > prev) && gap <= maxgap)
      { //cout << "  testing " << gap << " " << prev << " " << cur << " " << next << endl;
        int m0, m1;
        m0 = freq - kb;
        m1 = m0 + (int)(partial * gap * (sqrt(1 + inharm * (pow(partial, 2) - 1.0)) - 1));
        if(m1 > m0 + gap || m0 < 0)
        {
          m0 = 0;
          m1 = gap;
        }
        sum = 0;
        tmp[0] = curve[band][0];
        tmp[1] = curve[band][1];
        tmp[2] = curve[band][2];
        tmp[3] = curve[band][3];
        J = (int)floor((float)(Kb - m0 - 1) / gap) + 1;
        for(m = m0; J > 1 && sum != Lb[band][gap] && m <= m1; m++)
        {
          sum= 0.0;
          J = (int)floor((float)(Kb - m - 1) / gap) + 1;
          for(j = 0; j < J; j++) sum += Xe[band][m + (gap * j)];
          sum *= (.75/J + .25);
        }
        m += kb - 1;
        inharm_next = (pow(m / (partial * n), 2)-1) / (pow(partial,2)-1);
        if(sum != Lb[band][gap])
        {
          m = (int)floor(partial * (gap - 0.5));
          m1 = (int)ceil(partial * (gap + 0.5));
          if(m < lb - gap) m = lb - gap;
          if(m < kb) m = kb;
          if(m1 >= lb) m1 = lb - 1;
          while(m <= m1 && Lb[band][gap] != Xe[band][m-kb]) m++;
          J = 1;
          inharm_next = (pow((gap + 0.5)/n, 2) - 1) / (pow(partial, 2) - 1);
          sum = Xe[band][m-kb];
        }
        if(sum == Lb[band][gap])
        {
          curve[band][0] = partial;
          curve[band][1] = m;
          curve[band][2] = gap;
          curve[band][3] = J;
          sum += weight;
          if(band+1 == bins && sum > maxsum) maxsum = sum;
          if(band+1 != bins) sum = estimate(n,band+1,sum,inharm_next);
        }
        else sum = 0;
        if(sum != maxsum)
        {
          curve[band][0] = tmp[0];
          curve[band][1] = tmp[1];
          curve[band][2] = tmp[2];
          curve[band][3] = tmp[3];
        }
        if(sum > max) max = sum;
      }
      gap++;
      prev = cur;
      cur = next;
      next = Lb[band][gap + 1];
      //while(Lb[band][gap] < Lb[band][gap - 1] || Lb[band][gap] < Lb[band][gap + 1] && gap <= maxgap) gap++;
    }
    if(max == 0 && weight > maxsum)
    {
      max = weight;
      //cout << "max = " << max << endl;
      maxsum = max;
      b = band + 1;
      while(b < bins) curve[b++][0] = 0;      
    }
    return max;
}


//void
//Pitches::spectralSmooth(


float
Pitches::triangulate(int n, int Kb, int mid)
{
  if(n < mid) return (float)(n + 1) / (mid + 1);
  else return (float)(Kb - n)/(Kb - mid +1);
}


inline int
Pitches::round(float r)
{
  if(r - floor(r) <= 0.5 ) return (int)floor(r);
  return (int)ceil(r);
}


inline float
Pitches::freq(int n)
{
  return rate * n / size;
} 


inline int
Pitches::index(float f)
{
  return (int)(f * size / rate);
} 

	

void
Pitches::printFfts()
{
  float max, sum;
  int i, n, k;
  for(i = 0, n = start, max = 0; n < N; n++)
  {
    sum = 0;
    if(binStart[i] + binSz[i] <= n) i++;
    for(k = i; k < bins && binStart[k] <= n; k++) sum += Xe[k][n - binStart[k]];
    sum = sqrt(sum);
    ffts << n << " " << sum + offset << endl;
    if(sum > max) max = sum;
  }
  offset += max;
  ffts << endl;
}

void
Pitches::printWeights()  
{
  float max, sum;
  int i, n;
  for(i=0; i<bins; i++)
  {
    max = 0;
    for(n=start; n<N; n++)
    {
      sum = Lb[i][n];
      lbs << n << " " << sum + lboffset << endl;
      if(sum > max) max = sum;
    }
    lboffset += max;
    lbs << endl;
  }
  lboffset += max;
  lbs << endl; 
}


