/*
** Copyright (C) 2000 George Tzanetakis <gtzan@cs.princeton.edu>
**  
** This program is free software; you can redistribute it and/or modify
** it under the terms of the GNU General Public License as published by
** the Free Software Foundation; either version 2 of the License, or
** (at your option) any later version.
** 
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU General Public License for more details.
** 
** You should have received a copy of the GNU General Public License
** along with this program; if not, write to the Free Software 
** Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
*/


/**
   \class  SegmentorSortedPeaks
   \brief  Segment algorithm based on peak detection

   Constrain segmentation by desired number of peaks and acceptable 
peak spacing. The algorithm is described in a paper presented at 
WASPAA99 by the author and Perry Cook. The paper can be found in 
http::/www.cs.princeton.edu/~gtzan/papers. 
   Intuitively the signal is viewed as a trajectory of points corresponding 
to the feature vectors, in a high-dimensional space. The segmentation 
boundaries are found by detecting abrupt changes in this trajectory. 
More technically, after feature vectors are extracted, a distance signal 
is calculated by taking the Mahalanobis distance between succesive feature 
vectors. To further sharpen the peaks of the resulting distance signal the 
first derivative is taken. Finally the peaks of the distance signal correspond to the segment boundaries. The largest numPeaks_ peaks are retained as segmentation boundaries by succesive maximum calculation. Peak spacing is used 
to zero an area around the detected peak to avoid peaks that correspond 
to the same segment boundary. 
*/





#if !defined(__SegmentorSortedPeaks_h) 
#define __SegmentorSortedPeaks_h

#include "Segmentor.h"

class SegmentorSortedPeaks:public Segmentor
{
protected:
  string type_;
  unsigned int numPeaks_;
  unsigned int peakSpacing_;
  
public:
  SegmentorSortedPeaks();
  ~SegmentorSortedPeaks();
  void init(unsigned int numPeaks_, unsigned int peakSpacing_);
  void segment(FeatMatrix& fmatrix, fvec &segmentation);
  float mahalanobis(fvec& vec1, const fvec& vec2, const fvec& covars);
};

inline float 
SegmentorSortedPeaks::mahalanobis(fvec& vec1, const fvec& vec2, const fvec& covars)
{
  unsigned int i;
  float sq_sum = 0.0;
  unsigned int size = vec1.size();
  float diff;
  for (i=0; i<size; i++)
    {
      diff = vec1(i) - vec2(i);
      sq_sum += diff * covars(i) * diff;
    }
  return sq_sum;
}



#endif

   

