/*
** Copyright (C) 2000 George Tzanetakis <gtzan@cs.princeton.edu>
**  
** This program is free software; you can redistribute it and/or modify
** it under the terms of the GNU General Public License as published by
** the Free Software Foundation; either version 2 of the License, or
** (at your option) any later version.
** 
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU General Public License for more details.
** 
** You should have received a copy of the GNU General Public License
** along with this program; if not, write to the Free Software 
** Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
*/

/**
   \class SpectralSegm
   \brief FFT based Spectral features 


   Calculate FFT mangitude based features. 
The Centroid, Rolloff and Flux are calculated 
on the magnitude spectrum of the input fvec 
and Zcrs and RMS are calculated in time domain. 
This System (feature) is used for segmentation. 
*/



#include "SpectralSegm.h"

SpectralSegm::SpectralSegm()
{
  inSize_ = DEFAULT_WIN_SIZE;
  memSize_ = 10;
  outSize_ = 11;
  init();
}

SpectralSegm::SpectralSegm(MSignal *src, unsigned int memSize)
{
  inSize_ = src->winSize();
  memSize_ = memSize;
  outSize_ = 11;
  init();
}


void 
SpectralSegm::init()
{
  temp0_.create(3);  
  temp1_.create(1);
  temp2_.create(1);
  temp3_.create(10);
  temp4_.create(2);
  temp5_.create(1);
  final_.create(5);
  
  hamming_ = new Hamming(inSize_, 0);
  magfft_  = new MagFFT(inSize_);
  shapeDesc_  = new ShapeDesc(magfft_->outSize());
  zcrs_ = new Zcrs(inSize_);
  rms_ = new RMS(inSize_);
  memory_ = new Memory(outSize_/2, memSize_);
  lowMemory_ = new Memory(1, memSize_);
  vector<System *> featArray;
  featArray.push_back(hamming_);
  featArray.push_back(magfft_);
  featArray.push_back(shapeDesc_);
  series_ = new Series(featArray);
  
  
  featSize_ = 11;
  featNames_.push_back("MeanCentroid");
  featNames_.push_back("MeanRolloff");
  featNames_.push_back("MeanFlux");
  featNames_.push_back("MeanZerocrossings");
  featNames_.push_back("MeanRMS");
  featNames_.push_back("StdCentroid");
  featNames_.push_back("StdRolloff");
  featNames_.push_back("StdFlux");
  featNames_.push_back("StdZerocrossings");
  featNames_.push_back("StdRMS");
  featNames_.push_back("LowEnergy");
}



SpectralSegm::~SpectralSegm()
{
  delete hamming_;
  delete magfft_;
  delete shapeDesc_;
  delete zcrs_;
  delete rms_;
  delete memory_;
  delete lowMemory_;
  delete series_;
}


void 
SpectralSegm::process(fvec& in, fvec& out) 
{
  unsigned int i;
  assert((in.size() == inSize_) && (out.size() == outSize_));      
  
  series_->process(in, temp0_);
  zcrs_->process(in, temp1_);
  rms_->process(in, temp2_);
  
  for (i=0; i<3; i++)
    {
      final_(i) = temp0_(i);
    }
  final_(3) = temp1_(0);
  final_(4) = temp2_(0);
  memory_->process(final_, temp3_);
  lowMemory_->process(temp2_, temp4_);
  lowMemory_->getPerc(0.50, temp5_);
  for (i=0; i < temp3_.size(); i++)
    {
      out(i) = temp3_(i);
    }
  out(temp3_.size()) = temp5_(0);
}


	

	
