/*
** Copyright (C) 2000 George Tzanetakis <gtzan@cs.princeton.edu>
**  
** This program is free software; you can redistribute it and/or modify
** it under the terms of the GNU General Public License as published by
** the Free Software Foundation; either version 2 of the License, or
** (at your option) any later version.
** 
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU General Public License for more details.
** 
** You should have received a copy of the GNU General Public License
** along with this program; if not, write to the Free Software 
** Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
*/


/** 
    \class imatrix
    \brief Matrix of int values

    Matrix of int values. Similar to ivec. Basic 
arithmetic operations and statistics are supported. 
*/


#include "imatrix.h"

imatrix::imatrix(): data_(NULL), size_(0), rows_(0), cols_(0), name_("m"), printHeader_(true)
{
}

imatrix::~imatrix()
{
  delete data_;
}

imatrix::imatrix(unsigned int rows): data_(new int[rows * rows]), size_(rows*rows), rows_(rows), cols_(rows), name_("m"), printHeader_(true)
{
  row_.create(cols_);
  col_.create(rows_);
}


imatrix::imatrix(unsigned int rows, unsigned int cols): data_(new int[rows * cols]), size_(rows*cols), rows_(rows), cols_(cols), name_("m"), printHeader_(true)
{
  row_.create(cols_);
  col_.create(rows_);  
}


imatrix::imatrix(const imatrix& a):data_(new int[a.size_]), size_(a.size_), rows_(a.rows_), cols_(a.cols_), name_("m"), printHeader_(a.printHeader_)
{
  for (unsigned int i=0; i<size_; i++)
    data_[i] = a.data_[i];
  row_.create(cols_);
  col_.create(rows_);  
}
					 



imatrix&
imatrix::operator=(const imatrix& a)
{
  if (this != &a)
    {
      if ((rows_ != a.rows_)|| (cols_ != a.cols_))
	{
	  cerr << "imatrix::operator=: Different imatrix dimensions\n" << endl;
	  cerr << "imatrix left unchanged\n";
	  cerr << "Left rows = " << rows_ << endl;
	  cerr << "Left cols = " << cols_ << endl;
	  cerr << "Right rows = " << rows_ << endl;
	  cerr << "Right cols = " << cols_ << endl;
	  return *this;
	}
      for (unsigned int i=0; i< size_; i++)
	{
	  data_[i] = a.data_[i];
	}
      name_ = a.name_;
      printHeader_ = a.printHeader_;
    }
  return *this;
}


void 
imatrix::create(unsigned int rows)
{
  rows_ = rows;
  cols_ = rows;
  size_ = rows_ * cols_;
  data_ = new int[size_];
  for (unsigned int i=0; i<size_; i++)
    data_[i] = 0;
  row_.create(cols_);
  col_.create(rows_);
}

void 
imatrix::create(unsigned int rows, unsigned int cols)
{
  unsigned int i;
  rows_ = rows;
  cols_ = cols;
  size_ = rows_ * cols_;
  
  data_ = new int[size_];
  for (i=0; i<size_; i++)
    {
      data_[i] = 0;
    }
  row_.create(cols_);
  col_.create(rows_);
}


void 
imatrix::setval(int val)
{
  unsigned int r,c;
  for (r=0; r < rows_; r++)
    for (c=0; c < cols_; c++)
      data_[r*cols_ + c] = val;
}

unsigned int 
imatrix::rows()
{
  return rows_;
}

unsigned int 
imatrix::cols()
{
  return cols_;
}


void 
imatrix::col(const unsigned int c, ivec& vec)
{
  cerr << "imatrix::col c = " << c << endl;
  if (c >= cols_)
    {
      cerr <<"imatrix::col Column out of range\n" << endl;
      return;
    }
  if (vec.size() != rows_)
    {
      cerr << "imatrix::col Incompatible vector size\n" << endl;
      return;
    }
  
  for (unsigned int r=0; r<rows_; r++)
    {
      data_[r*cols_ +c] = vec(r);
    }
}



void 
imatrix::row(const unsigned int r, ivec& vec)
{
  if (r >= rows_)
    {
      cerr <<"imatrix::row(const int r, ivec& vec)  Row out of range ( " << r << ") \n" << endl;
      return;
    }
  if (vec.size() != cols_)
    {
      cerr << "imatrix::row Incompatible vector size\n" << endl;
      return;
    }
  for (unsigned int c=0; c<cols_; c++)
    {
      data_[r*cols_ +c] = vec(c);
    }
}


imatrix& imatrix::operator*=(const int val)
{
  unsigned int r, c;
  
  
  for (r=0; r < rows_; r++)
    {
      for (c=0; c < cols_; c++)
	{
	  data_[r * cols_ + c] *= val;
	}
    }
  return *this;
}



imatrix& imatrix::operator/=(const int val)
{
  unsigned int r, c;
  
  
  for (r=0; r < rows_; r++)
    {
      for (c=0; c < cols_; c++)
	{
	  data_[r * cols_ + c] /= val;
	}
    }
  return *this;
}



imatrix& imatrix::operator+=(const imatrix& matrix)
{
  unsigned int r, c;
  
  if ((matrix.rows_ != cols_)||(matrix.cols_ != cols_))
    {
      cerr << "imatrix::operator+=: Wrong # of columns" << endl;
      cerr << "imatrix left unchanged" << endl;
      return *this;
    }
  for (r=0; r < rows_; r++)
    {
      for (c=0; c < cols_; c++)
	{
	  data_[r * cols_ + c] += matrix(r,c);
	}
    }
  return *this;
}


imatrix& imatrix::operator*=(const imatrix& matrix)
{
  unsigned int r, c, j;
  int sum;
  imatrix temp(rows_, cols_);
  for (r=0; r < rows_; r++)
    for (c=0; c < cols_; c++)
      temp(r,c) = data_[r * cols_ + c];
  
  if ((matrix.rows_ != cols_)||(matrix.cols_ != cols_))
    {
      cerr << "imatrix::operator*=: Wrong # of columns" << endl;
      cerr << "imatrix left unchanged" << endl;
      return *this;
    }
  for (r=0; r < rows_; r++)
    {
      for (c=0; c < cols_; c++)
	{
	  sum = 0;
	  for (j=0; j < cols_; j++)
	    sum += ((temp(r ,j))  * matrix(j,c));
	  data_[r*cols_ + c] = sum;
	}
    }
  return *this;
}




ivec
imatrix::col(const unsigned int c) 
{
  if (c >= cols_)
    {
      cerr <<"imatrix::col Column out of range\n" << endl;
      col_.setval(0);
      return col_;
    }

  for (unsigned int r=0; r<rows_; r++)
    {
      col_(r) = data_[r*cols_ +c];
    }
  return col_;
}

ivec
imatrix::row(const unsigned int r) 
{
  if (r >= rows_)
    {
      cerr << "imatrix::row(const long r) Row out of range (" << r << ") rows_ =  " << rows_ << "\n" << endl;
      row_.setval(0);
      return row_;
    }      
  for (unsigned long c=0; c<cols_; c++)
    {
      row_(c) = data_[r*cols_ +c];
    }
  return row_;
}



void 
imatrix::read(string filename)
{
  ifstream from(filename.c_str());
  from >> (*this);
}



istream& operator>> (istream& is, imatrix& fm)
{
  unsigned int i,r,c;
  unsigned int rows, cols;
  string str0, str1, str2;
  is >> str0 >> str1;
  
  
  if ((str0 != "#") || (str1 != "name:"))
    {
      cerr << "imatrix:::operator>>: Problem reading fmatrix object from istream" << endl;
      cerr << "str0 = " << str0 << endl;
      return is;
    }
  is >> fm.name_;
  is >> str0 >> str1 >> str2;
  is >> str0 >> str1;
  is >> rows;
  is >> str0 >> str1;
  is >> cols;
  
  fm.create(rows, cols);
  for (r=0; r< rows; r++)
    {
      for (c=0; c < cols; c++)
	{
	  is >> fm(r,c);
	}
    }
  return is;
}

void 
imatrix::printHeader(bool flag)
{
  printHeader_ = flag;
}


ostream& operator<< (ostream& o, const imatrix& m)
{
  int rows, cols;
  rows = m.rows_;
  cols = m.cols_;
  
  int r,c;
  if (m.printHeader_)
    {
      o << "# name: " << m.name_ << endl;
      o << "# type: matrix" << endl;
      o << "# rows: " << rows << endl;
      o << "# columns: " << cols << endl << endl;
    }
  for (r=0; r < rows; r++)
    {
      for (c=0; c < cols; c++)
	{
	  o << m(r,c) << "\t";
	}
      o << endl;
    }
  return o;
}
 
