/*
** Copyright (C) 2000 George Tzanetakis <gtzan@cs.princeton.edu>
**  
** This program is free software; you can redistribute it and/or modify
** it under the terms of the GNU General Public License as published by
** the Free Software Foundation; either version 2 of the License, or
** (at your option) any later version.
** 
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU General Public License for more details.
** 
** You should have received a copy of the GNU General Public License
** along with this program; if not, write to the Free Software 
** Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
*/


/** 
    \class ivec
    \brief Vector of int values


    Array (vector in the numerical sense) of int values. Basic 
arithmetic operations and statistics are supported. 
*/


#include "ivec.h"


ivec::ivec()
{
  size_ = 0;
  data_ = NULL;
}


ivec::~ivec()
{
  delete data_;
}

ivec::ivec(unsigned int size)
{
  size_ = size;
  data_ = new int[size_];
}

ivec::ivec(const ivec& a):data_(new int[a.size_]), size_(a.size_)
{
  for (unsigned int i=0; i<size_; i++)
    data_[i] = a.data_[i];
}

      


ivec& 
ivec::operator=(const ivec& a)
{
  if (this != &a)
    {
      if (size_ != a.size_)
	{
	  cerr << "ivec::operator= : Different ivec sizes\n" << endl;
	  cerr << "ivec left unchanged\n";
	  cerr << "Left size = " << size_ << endl;
	  cerr << "Right size = " << a.size_ << endl;
	  return *this;
	}
      for (unsigned int i=0; i < size_; i++)
	data_[i] = a.data_[i];
    }
  return *this;
}



int *
ivec::get_data()
{
  return data_;
}

float 
ivec::mean()
{
  float sum = 0.0;
  for (unsigned int i=0; i < size_; i++)
    {
      sum += data_[i];
    }
  if (sum != 0.0) 
    sum /= size_;
  return sum;
}

float 
ivec::var()
{
  float sum = 0.0;
  float sum_sq = 0.0;
  int val;
  float var;
  
  for (unsigned int i=0; i < size_; i++)
    {
      val = data_[i];
      sum += val;
      sum_sq += (val * val);
    }
  if (sum != 0.0) sum /= size_;
  if (sum_sq != 0.0) sum_sq /= size_;
  
  var = sum_sq - sum * sum;
  return var;
}

float 
ivec::std()
{
  return sqrt(var());
}

  

unsigned int
ivec::size()
{
  return size_;
}

void 
ivec::debug_info()
{
  cerr << "ivec information" << endl;
  cerr << "size = " << size_ << endl;
  cerr << "****************" << endl;
  for (unsigned int i=0; i<size_; i++)
    cerr << data_[i] << endl;
  cerr << "****************" << endl;
}


void 
ivec::create(unsigned long size)
{
  size_ = size;
  data_ = new int[size_];
  for (unsigned long i=0; i<size_; i++)
    data_[i] = 0;
}


void 
ivec::create(int val, unsigned long size)
{
  size_ = size;
  data_ = new int[size_];
  for (unsigned long i=0; i<size_; i++)
    data_[i] = val;
}


void 
ivec::allocate(unsigned long size)
{
  size_ = size;
  data_ = new int[size_];
}

void
ivec::setval(unsigned int start, unsigned int end, int val)
{
  assert(start <= size_);
  assert(end <= size_);
  for (unsigned int i=start; i<end; i++)
    {
      data_[i] = val;
    }
}


void 
ivec::setval(int val)
{
  for (unsigned int i=0; i<size_; i++)
    {
      data_[i] = val;
    }
}


void
ivec::sqr()
{
  for (unsigned int i=0; i<size_; i++)
    {
      data_[i] *= data_[i];
    }
}




ivec 
operator+(const ivec& vec1, const ivec& vec2)
{
  unsigned int size;
  unsigned int i;
  if (vec1.size_ != vec2.size_)
    cerr << "Size of ivecs does not match" << endl;
  if (vec1.size_ >= vec2.size_)
    size = vec1.size_;
  else 
    size = vec2.size_;
  ivec sum;
  sum.create(size);    
  
  for (i=0; i<vec1.size_; i++)
    {
      sum.data_[i] = vec1.data_[i];
    }
  for (i=0; i<vec2.size_; i++)
    {
      sum.data_[i] += vec2.data_[i];
    }
      
  return sum;
}


ivec 
operator-(const ivec& vec1, const ivec& vec2)
{
  unsigned int size;
  unsigned int i;
  if (vec1.size_ != vec2.size_)
    cerr << "Size of ivecs does not match" << endl;
  if (vec1.size_ >= vec2.size_)
    size = vec1.size_;
  else 
    size = vec2.size_;
  ivec diff;
  diff.create(size);    
  
  for (i=0; i<vec1.size_; i++)
    {
      diff.data_[i] = vec1.data_[i];
    }
  for (i=0; i<vec2.size_; i++)
    {
      diff.data_[i] -= vec2.data_[i];
    }
      
  return diff;
}


void 
ivec::send(Communicator *com)
{ 
  unsigned int i;
  static char *buf = new char[256];
  string message;
  sprintf(buf, "%i\n", size_);
  message = buf;
  com->send_message(message);
  for (i=0; i<size_; i++)
    {
      sprintf(buf, "%d\n", data_[i]);
      message = buf;
      com->send_message(message);
    }
  cerr << "ivec::send " << size_ << " numbers were sent to the client\n";
}


ostream& 
operator<< (ostream& o, const ivec& vec)
{
  o << "# MARSYAS ivec" << endl;
  o << "# Size = " << vec.size_ << endl << endl;
  o << endl;

  
  o << "# name: v" << endl;
  o << "# type: matrix" << endl;
  o << "# rows: " << vec.size_ << endl;
  o << "# columns: " << 1 << endl;


  for (unsigned int i=0; i<vec.size_; i++)
    o << vec.data_[i] << endl;
  o << endl;
  o << "# Size = " << vec.size_ << endl;
  o << "# MARSYAS ivec" << endl;  
  return o;
}

istream& 
operator>>(istream& is, ivec& vec)
{
  if (vec.size_ != 0)
    {
      cerr << "ivec::operator>>: ivec already allocated cannot read from istream\n" << endl;
      return is;
    }
  string str0,str1,str2;
  unsigned int size;
  unsigned int i;
  
  is >> str0;
  is >> str1;
  is >> str2;
  
  if ((str0 != "#") || (str1 != "MARSYAS") || (str2 != "ivec"))
    {
      cerr << "ivec::operator>>: Problem reading ivec object from istream" << endl;
      return is;
    }
  is >> str0;				
  is >> str1;
  is >> str2;
  if ((str0 != "#") || (str1 != "Size") || (str2 != "="))
    {
      cerr << "ivec::operator>>: Problem reading ivec object from istream" << endl;
      return is;
    }
  is >> size;
  
  for (i=0; i<12; i++)
    {
      is >> str0;
    }
  
  
  vec.allocate(size);
  for (i=0; i < size; i++)
    {
      is >> vec.data_[i];
    }
  is >> str0;				
  is >> str1;
  is >> str2;
  if ((str0 != "#") || (str1 != "Size") || (str2 != "="))
    {
      cerr << "ivec::operator>>: Problem reading ivec object from istream" << endl;
      return is;
    }
  is >> size;
  
  is >> str0;
  is >> str1;
  is >> str2;
  
  if ((str0 != "#") || (str1 != "MARSYAS") || (str2 != "ivec"))
    {
      cerr << "ivec::operator>>: Problem reading ivec object from istream" << endl;
      return is;
    }
  return is;
}
