/*
** Copyright (C) 2000 George Tzanetakis <gtzan@cs.princeton.edu>
**  
** This program is free software; you can redistribute it and/or modify
** it under the terms of the GNU General Public License as published by
** the Free Software Foundation; either version 2 of the License, or
** (at your option) any later version.
** 
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU General Public License for more details.
** 
** You should have received a copy of the GNU General Public License
** along with this program; if not, write to the Free Software 
** Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
*/

#include <stdio.h>
#include <stdlib.h>
#include <sys/stat.h>
#include "FeatMatrix.h"
#include "BatchExtract.h"
#include "MSignal.h"
#include "fmatrix.h"
#include "SoundFileIO.h"
#include "defs.h"
#include "Util.h"
#include "Gain.h"
#include "Hamming.h"
#include "MagFFT.h"
#include "FeatExtractor.h"
#include "Series.h"
#include "MagDB.h"
#include "Sine.h"
#include "Spectral.h"
#include "Collection.h"
#include "SpectralSegm.h"
#include "FeatExtractor.h"
#include "Shuffler.h" 
#include "CollectionShuffler.h"
#include "MFCC.h"
#include "ExtractorDictionary.h"
#include "Classifier.h"
#include "MemMFCC.h"
#include "LPC.h"
#include "SfxExtractor.h"
#include "DWTC.h"
#include "BeatExtractor.h"
#include "SvFeatExtractor.h"
#include "MultiPitchExtractor.h"
#include "GenreExtractor.h"
#include "MultiPitchFeature.h"
#include "Parallel.h"
#include "Series.h"
#include "ValleyPick.h"
#include "AMDF.h"
#include "knnTree.h"
#include "FFT.h"
#include <map>

#define PITCHES 1 

struct ltstr
{
  bool operator()(const char* s1, const char* s2) const
  {
    return strcmp(s1, s2) < 0;
  }
};

inline void
mapNns(ivec& mapping, int **nns, FeatMatrix* templRes, FeatMatrix* collectRes, int strings);

FeatMatrix*
stringify(int nExtract, FeatMatrix** fMatrixes, int strings);


int
main(int argc, char **argv)
{
    

  if (argc < 8)
  {
  	cerr << "Usage :" << argv[0] <<" windowSize hopSize template"; 
    cerr << " templateGain collection collectGain [extractor weight]...\n"; 
    cerr << "  windowSize:     number of samples. ie 512, 4096\n"; 
    cerr << "                    (Must be a power of 2)\n";
    cerr << "  hopSize:        amount window shifts ie 256, 512\n";
    cerr << "                    (must be a power of 2)\n";
    cerr << "  template:       audiofile that is basis of mosaic\n";
    cerr << "  templateGain:   volume of template  in output\n";
    cerr << "                    (must be between 0.0 & 1.0)\n";
    cerr << "  collection:     collection used to replace template sound\n";
    cerr << "  collectGain:    volume of collection in output\n";
    cerr << "                    (must be between 0.0 & 1.0)\n";
    cerr << "  strings:        maximum number of concurrent matching windows\n";
    cerr << "  extractor wt:   SFFT weight\n";
    cerr << "                  PITCH weight\n";
    cerr << "                  MFCC weight\n";
    exit(1);  
  }
  
  
  /*parse input*/
  
  unsigned int winSize = atoi(argv[1]);
  unsigned int hopSize = atoi(argv[2]);
  string templName     = argv[3];
  double templGain     = atof(argv[4]);
  string collectName   = argv[5]; 
  double collectGain   = atof(argv[6]);
  unsigned int strings = atoi(argv[7]);
  string winSizeStr    = argv[1];
  string hopSizeStr    = argv[2];

  unsigned int nExtractors = (argc - 8)/2;
  string extractors[nExtractors];
  string extractorWtsStr[nExtractors];
  float extractorWts[nExtractors];
  unsigned int i;
  string nTreeFile = collectName;
  for(i = 0; i < nExtractors; i++)
  {
    extractors[i] = argv[i*2 + 8];
    extractorWts[i] = atof(argv[i*2 + 9]);
    nTreeFile += extractors[i];
    nTreeFile += argv[i*2 + 9]; 
    nTreeFile += "-"; 
  }
  nTreeFile += winSizeStr;
  nTreeFile += "-";
  nTreeFile += hopSizeStr;
  nTreeFile += "-";
 
  cout << "windowSize  = " << winSize << endl;
  cout << "hopSize     = " << hopSize << endl;
  cout << "template    = " << templName << ": " << templGain << endl;
  cout << "collection  = " << collectName << ": " << collectGain << endl;
  cout << "extractors  = " << extractors[0]<< ": " <<extractorWts[0]<<endl;
  cout << "strings     = " << strings << endl;
  for(i = 1; i < nExtractors; i++)
    cout << "              "<<extractors[i]<<" : "<<extractorWts[i]<<endl;



  /*retrieve template*/

  struct stat stbuf;
  string extractor = extractors[0]; 
  if(stat(templName.c_str(), &stbuf))
  {
    cout << "shuffle: template '" << templName << "' does not exist." << endl;
    exit(1);
  }

  Util util;
  MSignal *templ;
  if((templ = util.sfopen(templName, MRS_SF_READ)) == NULL)
  {
    cout << "shuffle: template '" << templName << "' does not exist" << endl;
    exit(1);
  }
  templ->initWindow(winSize, hopSize, 0, 0);
   cout << templ->srate() << endl;
 
  
  /*get FeatMatrixes for template and collection for each feature*/
  
  ExtractorDictionary *dict;
  dict = new ExtractorDictionary(templ);
  Extractor *extr;
  int* startnum; 
  startnum = NULL; 
  unsigned int featSizes[nExtractors];
  BatchExtract *collectExtract;
  collectExtract = new BatchExtract(collectName, winSizeStr, hopSizeStr);
  FeatMatrix* collectReses[nExtractors];
  FeatMatrix* templReses[nExtractors];

  for(i = 0; i < nExtractors; i++) 
  {
    extr = dict->getExtractor(extractors[i]);
    templReses[i] = dict->getFeatMatrix(extractors[i]);
    extr->extract(*(templReses[i]));
    collectReses[i] = collectExtract->extract(extractors[i], &startnum);
    featSizes[i] = collectReses[i]->cols();
  }
  
 
  /*combine Featmatrixes for multiple features into collectRes and templRes*/
  
  unsigned int k;
  int labelIndexes[collectReses[0]->rows()];
  for(i = 0; i < collectReses[0]->rows(); i++) 	
    labelIndexes[i] = collectReses[0]->labelNum(i, 1);
  FeatMatrix *templRes, *collectRes;
  templRes = stringify(nExtractors, templReses, strings);
  collectRes = stringify(nExtractors, collectReses, strings);

  delete dict;
  delete collectExtract;

  /*multiply features by weights*/

  unsigned int j, dim = collectRes->cols();
  fvec covariance(dim);
  fvec vars(dim);
  vars = templRes->varRow();  
  fvec templVec(templRes->rows());
  fvec collectVec(collectRes->rows());
  
  for (j = 0; j < dim; j++)
  {
    if (vars(j) <= 0.000001)  covariance(j) = 1.0;
    else  covariance(j) = 1.0 / vars(j);
  }
  
  for(i = 0, j = 0; i < nExtractors; i++)
  {
     unsigned int featSize = strings * featSizes[i]; 
     for(k = 0; k < featSize; k++, j++)
     {  
        templVec = templRes->col(j);
        collectVec = collectRes->col(j);
        templVec *= (extractorWts[i] * sqrt(covariance(j)));
        collectVec *= (extractorWts[i] * sqrt(covariance(j)));
        templRes->col(j, templVec);
        collectRes->col(j, collectVec);
     }
  }



  /*build search tree or read in if it already exists*/

  int n_pts=collectRes->rows();
  knnTree tree(collectRes, n_pts, dim, 7);

  string treeFile = MARSYAS_MFDIR;
  treeFile += "/trees/";
  treeFile += nTreeFile;
  char stringsStr[2];
  stringsStr[1] = '\0';

  if(strings >= 10)
  {
    stringsStr[0] = (char)(strings/10 + (int)'0');
    treeFile += stringsStr;
  }
  stringsStr[0] = (char)(strings%10 + (int)'0');
  treeFile += stringsStr;
  treeFile += ".mtf";
 
  if(stat(treeFile.c_str(), &stbuf))
  {
    cout << "Building search tree..." << endl;
    tree.makeTree();
    cout << "Writing search tree..." << endl;
    tree.write(treeFile);
  }
  else
  {
    cout << "Reading search tree..." << endl;
    tree.read(treeFile);
  }  



  /*search for nearest neighbor string*/

  unsigned int n = templRes->rows() + strings - 1;  
  ivec mapping(n);
  int *nns[n];
  float *dist[n];
  fvec q(dim);
  unsigned int percent=0;
  cout << "0%" << endl;
  for(i=0; i<n; i++)
  {
    if((i*100)/n > percent) { percent=(i*100)/n; cout<<percent<<"%"<<endl; }
    q=templRes->row(i);
    nns[i] = new int[1];
    dist[i] = new float[1];
    tree.findKnn(q, 1, nns[i], dist[i]);
  }

  cout << "100%" << endl;  

  mapNns(mapping, nns, templRes, collectRes, strings);

  cout << mapping << endl;
 
  

  /*prepare output*/
 
  AuFileIO *dest;
  dest = new AuFileIO("shuffled.au", MRS_SF_WRITE);
  dest->init();
  dest->initWindow(templ->hopSize(), templ->hopSize(), 0, 
    templ->zeroSize());
  dest->initMSignal(44100, 1);



  /*create soundfile based on mapping*/

  fvec outwindow1(hopSize);
  outwindow1.setval(0.0);
  fvec overlap(winSize-hopSize);
  overlap.setval(0.0);
  fvec outwindow2(hopSize);
  outwindow2.setval(0.0);
  fvec window1(templ->winSize());
  fvec window2(templ->winSize());
  fvec collFft(templ->winSize() + 2);
  fvec templFft(templ->winSize() + 2);
  RMS rms(winSize);
  fvec rms1(1);
  fvec rms2(1);
  float factor, overlapfSize = winSize - hopSize;
  unsigned int label, overlapSize = winSize - hopSize;
  MSignal *sample;

  FFT fftt(winSize, 1, 1, templ->srate());
  FFT fftc(winSize, 1, 1, templ->srate());

  for (i=0; i < mapping.size(); i++)
    {
      label = labelIndexes[mapping(i)];            
      sample = util.sfopen(collectRes->labelNames()[label+1], MRS_SF_READ);
      sample->initWindow(winSize,hopSize,0,0);
      sample->get(window1, mapping(i)-startnum[label]);
      templ->get(window2, i);
      delete sample;

      rms.process(window1, rms1);
      rms.process(window2, rms2);
      factor = collectGain * rms2(0) / rms1(0);
      
      fftt.process(window2, templFft);
      fftc.process(window1, collFft);
      for(j = 0; j < winSize; j+=2)
      {
        if(templFft(j) == 0.0)
        {
          templFft(j) = collFft(j);
          templFft(j+1) = collFft(j+1);
        }
        else templFft(j) -= collFft(j);
      }
      fftt.unprocess(templFft, window2);
      rms.process(window2, rms1);    
 
      for (j=0; j < overlapSize; j++)            
      {
      	outwindow1(j) = ((float)j/overlapfSize) * factor * 
      	                window1(j) + overlap(j);
      	overlap(j) = ((overlapfSize-(float)j)/overlapfSize) * factor *
                        window1(j+hopSize);
        outwindow2(j) = window2(j) * rms2(0)/rms1(0);
      }  
      for (j=overlapSize; j < hopSize; j++)
      {
        outwindow1(j) = window1(j) * factor;
        outwindow2(j) = window2(j);
      }
      dest->put(outwindow1, i);
    }

  return 0;
}



inline void
mapNns(ivec& mapping, int **nns, FeatMatrix* templRes, FeatMatrix* collectRes, int strings)
{
  int i, j, k, last=0, templSize=templRes->rows(), feats=collectRes->cols();
  fvec templVec(feats);
  fvec collectVec(feats);
  feats /= strings;
  float diff, dist, oldDist = 0, dists[strings], oldDists[strings-1]; 
 
  for(i = 0; i < templSize; i++)
  {
    templVec = templRes->row(i);
    collectVec = collectRes->row(nns[i][0]);
    for(j = 0; j < strings; j++)
    {
      dists[j] = 0.0;
      for(k = 0; k < feats; k++) 
      {
        diff = templVec(j*feats + k) - collectVec(j*feats + k);
        dists[j] += (diff * diff);
      }
    } 

    dist = 0.0;
    for(j = 0; j < last; j++) dist += dists[j];
    if(last == 0 || dist < oldDist)
    { 
      oldDist = dist - dists[0];
      for(k = last; k < strings - 1; k++) oldDist += dists[k];
      for(k = 0; k < strings; k++) mapping(i+k) = nns[i][0] + k;
      last = strings - 1;
      for(k = 0; k < last; k++) oldDists[k] = dists[k+1];
    }
    else
    {
      oldDist -= oldDists[0];
      last--;
      for(k = 0; k < last; k++) oldDists[k] = oldDists[k+1];
    }
  }
}


FeatMatrix *stringify(int nExtract, FeatMatrix** fMatrixes, int strings)
{
  int i, j, k, current;
  int nRows = fMatrixes[0]->rows() - strings + 1;
  int nCols = 0;
  for(i = 0; i < nExtract; i++) nCols += fMatrixes[i]->cols();
  nCols *= strings;

  FeatMatrix *newMatrix = new FeatMatrix(nRows, nCols);
  newMatrix->setLabelNames(fMatrixes[0]->labelNames());
  fvec newCol(nRows);
  for(i = 0, current = 0; i < nExtract; i++)
  {
    int feats = fMatrixes[i]->cols();
    float *data = fMatrixes[i]->getData();
    for(j = 0; j < feats*strings; j++, current++)
    {
      for(k = 0; k < nRows; k++) newCol(k) = data[k*feats + j];
      newMatrix->col(current, newCol);  
    }
  }
  return newMatrix;
}



